//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: action.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/action.cc,v $
//  
//  File	      : action.cc
//  Author(s)	      : Jaeho Lee <jaeho@coral.engin.umich.edu>
//  
//  Description	      : Actions
//  
//  Original author(s): Jaeho Lee <jaeho@coral.engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sun Mar 19 12:42:53 1995
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:05:43 1996
//  Update Count      : 73
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include "action.h"
#include "primitive.h"
#include "wm.h"
#include "goal.h"
#include "parse.h"


Action::ActStatus Primitive::execute(Binding* b)
{
  return function(arity, args, b).eval(b).is_true() ? ACT_SUCCEEDED : ACT_FAILED;
}


Action::ActStatus Load::execute(Binding* b)
{
  long error_count = 0;
    
  if (args->get_count() <= 0) {
    error_count += parse(0);
  }
  else {
    ExpListIterator next_exp(args);
    Expression* exp;

    for(exp = next_exp(); exp; exp = next_exp())
      error_count += parse((char*) exp->eval(b));
  }
  
  return error_count ? ACT_FAILED : ACT_SUCCEEDED;
}


Action::ActStatus Parse::execute(Binding* b)
{
  long error_count = 0;
    
  if (args->get_count() > 0) {
    Value val = args->first()->eval(b);

    if (val.type() == Value::VAL_STRING)
      error_count += parse_string(val);
    else
      return ACT_FAILED;
  }
  
  return error_count ? ACT_FAILED : ACT_SUCCEEDED;
}


Action::ActStatus Assign::execute(Binding* b)
{
  Value v = exp->eval(b);
  b->set_value(var, v);
  return ACT_SUCCEEDED;
}


Action::ActStatus Fact::execute(Binding* b)
{
  return (program.world_model->match(relation, b)) ? ACT_SUCCEEDED : ACT_FAILED;
}


Action::ActStatus Fail::execute(Binding* b)
{
  return ACT_FAILED;
}


Action::ActStatus Retrieve::execute(Binding* b)	
{ 
  // unbound variables before matching
  b->unbind_variables(relation->get_args());
    
  // now try match to get new values
  return (program.world_model->match(relation, b)) ? ACT_SUCCEEDED : ACT_FAILED; 
}


Action::ActStatus Test::execute(Binding* b)
{
  return exp->eval(b).is_true() ? ACT_SUCCEEDED : ACT_FAILED;
}


Action::ActStatus Assert::execute(Binding* b)
{
  program.world_model->assert(relation, b);
  if (program.show_wm)
    program.world_model->print();
  return ACT_SUCCEEDED;
}


Action::ActStatus Retract::execute(Binding* b)
{
  program.world_model->retract(relation, b);
  if (program.show_wm)
    program.world_model->print();
  return ACT_SUCCEEDED;
}


Action::ActStatus Update::execute(Binding* b)	
{ 
  program.world_model->update(old_relation, new_relation, b);
  if (program.show_wm)
    program.world_model->print();
  return ACT_SUCCEEDED; 
}


Action::ActStatus Post::execute(Binding* b)	
{
  program.goal_list->add(new GoalAction(_goal_action, b));
  return ACT_SUCCEEDED;
}

Action::ActStatus UnPost::execute(Binding* b)	
{
  program.goal_list->drop(_goal_action, b);
  return ACT_SUCCEEDED;
}

//
// GoalAction
//
int GoalAction::is_eligible(const char* name, Binding* binding)
{
  register Expression* e;
  char* str;
    
  if (_by) {
    ExpListIterator next_exp(_by);
      
    while(e=next_exp()) {
      str = e->eval(binding);
      if (str && ::strcmp(name, str) == 0) return 1;
    }

    return 0;
  }

  if (_not_by) {
    ExpListIterator next_exp(_not_by);
      
    while(e=next_exp()) {
      str = e->eval(binding);
      if (str && ::strcmp(name, str) == 0) return 0;
    }

    return 1;
  }

  return 1;
}

void GoalAction::format_args(ostream& s, Binding* b, char* head, char* tail)
{
  s << head;
  
  _goal->format(s, b);

  if (_priority)
    s << " :PRIORITY " << _priority->eval(b);

  if (_by) {
    s << " :BY ";
    explist_format(_by, s, b);
  }

  if (_not_by) {
    s << " :NOT-BY ";
    explist_format(_not_by, s, b);
  }
  
  s << tail;
}

