//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: binding.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/binding.cc,v $
//  
//  File	      : binding.cc
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : variable binding
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:34:40 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:05:41 1996
//  Update Count      : 133
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include "extern.h"
#include "binding.h"
#include "exp-var.h"
#include "exp-val.h"
#include "symtab.h"

//

Binding::Binding(SymbolTable* st) 
{ 
  // local binding attributes
  symtab = st;
  new_wm_binding = 0;
  bvalues = new BindingValue[size=st->get_size()];
}


Binding::Binding(Binding* b) 
{
  if (b) {
    symtab = b->symtab; 
    new_wm_binding = b->new_wm_binding;
    bvalues = new BindingValue[size = b->size];
    for (register long i = 0; i < size; i++)
      bvalues[i] = b->bvalues[i];
  }
  else {
    symtab = 0;
    new_wm_binding = 0;
    size = 0;
    bvalues = 0;
  }
}

void Binding::operator=(Binding& b) 
{
  if (this == &b) return;

  symtab = b.symtab; 
  new_wm_binding = b.new_wm_binding;

  delete[] bvalues;
  bvalues = new BindingValue[size = b.size];
  for (register long i = 0; i < size; i++)
     bvalues[i] = b.bvalues[i];
}

void Binding::unbind_variable(Expression* expression)
{
  if (expression->is_variable())
    set_value(expression, Value::Undefined);
}


void Binding::unbind_variables(ExpList* expressions)
{
  ExpListIterator next_exp(expressions);
  Expression* e; 

  while (e = next_exp())
    unbind_variable(e);
}


void Binding::link_variables(Expression* var, Expression* ext_variable, Binding* ext_binding)
{
  BindingValue& bval = bvalues[var->get_variable()->get_id()];

  bval.external_variable = (Variable*) ext_variable->get_variable();
  bval.external_binding = ext_binding;
}


int Binding::is_local_binding(Expression* var) const
{ 
  return !var->is_variable() || is_local_binding(var->get_variable()->get_id());
}


int Binding::is_local_binding(long var_id) const
{ 
  return bvalues[var_id].external_variable == 0;
}


void Binding::set_value(Expression* var, Value val) 
{
  if (!var->is_variable()) {
    cerr << "\nArgument should be variable!\n";
    return;
  }
  set_value(var->get_variable()->get_id(), val);
}


void Binding::set_value(long var_id, Value val) 
{
  BindingValue& bval = bvalues[var_id];

  if (is_local_binding(var_id))
    bval.value = val;
  else
    bval.external_binding->set_value(bval.external_variable, val);
}


Value& Binding::get_value(const Expression*const var) const
{
  return get_value(var->get_variable()->get_id());
}


Value& Binding::get_value(long const var_id) const
{
  BindingValue& bval = bvalues[var_id];

  if (is_local_binding(var_id))
    return bval.value;
  else
    return bval.external_binding->get_value(bval.external_variable);
}

//
int Binding::is_new_wm_binding() const
{
  return new_wm_binding;
}
  
int Binding::check_new_wm_binding(int new_wm)
{
  return new_wm_binding = new_wm_binding || new_wm;
}

void Binding::clear_new_wm_binding()
{
  new_wm_binding = 0;
}

//

void Binding::print(ostream& s)
{ 
  for (long i = 0; i < size; i++) {
    s << form("%3d: %s = ", i, symtab->lookup(i)->get_name());
    if (get_value(i).is_defined())
      s << get_value(i);
    else
      s << "NOT BOUND";
    s << "\n";
  }
}
