//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: binding.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/binding.h,v $
//  
//  File	      : binding.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : variable binding
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:34:18 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:37 1996
//  Update Count      : 142
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __BINDING_H__
#define __BINDING_H__

#include "dlist.h"
#include "exp-val.h"

//
class SymbolTable;
class Binding;
class Variable;

//
class BindingValue
{
friend class Binding;

protected:
  // Binding value can be either a constant value
  Value		value;

  // or a variable defined in other external binding
  Variable*	external_variable;
  Binding*	external_binding;

public:

  BindingValue()
  : value(Value::Undefined)

  {
    external_variable = 0; 
    external_binding = 0; 
  }

  BindingValue(Value& val)
  : value(val)
  { 
    external_variable = 0; 
    external_binding = 0; 
  }

  BindingValue(Variable* var, Binding* b)
  : value(Value::Undefined)
  { 
    external_variable = var;  
    external_binding = b;
  }
};

//

class Binding
{
  long		size;
  int		new_wm_binding;	// true if the binding is done with at least one new WM entry
  BindingValue*	bvalues;	// an array of values (variable-id is the index)
  SymbolTable*	symtab;

public:

  ~Binding()
  {
    delete[] bvalues;
  }

  Binding(SymbolTable* st);

  Binding(Binding* b);

  void operator=(Binding& b);

  //

  void		unbind_variable(Expression* expression);
  void		unbind_variables(ExpList* expressions);
  void		link_variables(Expression* var, Expression* ext_variable, Binding* ext_binding);

  int		is_local_binding(Expression* var) const; 
  int		is_local_binding(long var_id) const;

  void		set_value(Expression* var, Value val);
  void		set_value(long var_id, Value val);

  Value&	get_value(const Expression*const var) const;
  Value&	get_value(long const var_id) const;

  //
  
  int		is_new_wm_binding() const;
  int		check_new_wm_binding(int new_wm);
  void		clear_new_wm_binding();

//

  SymbolTable*	get_symtab() const	{ return symtab; }
  int		is_empty() const	{ return size == 0; }

  void		print(ostream& s=cout);
};

// Binding List
typedef Dlist<Binding*>		BindingList;
typedef DlistIterator<Binding*> BindingListIterator;

#endif




