//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: condition.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/condition.cc,v $
//  
//  File	      : condition.cc
//  Author(s)	      : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  
//  Description	      : condtions
//  
//  Original author(s): Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Wed Feb	 9 22:23:25 1994
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:05:37 1996
//  Update Count      : 153
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>

#include "extern.h"
#include "relation.h"
#include "condition.h"

//
// Expression Condition
// 

int ExpCondition::check(BindingList& bl) const
// remove from the given binding list the ones not satisfying the expression.
{
  BindingListIterator binding_iterator(bl);
  Binding* b;

  while (b = binding_iterator()) {

    if (!expression->eval(b).is_true()) {
      binding_iterator.remove_this();
      delete b;
    }
  }
  
  return bl.get_count();
}

int ExpCondition::confirm(Binding* b) const
// confirm whether the binding is still valid against the current WM
{
  return expression->eval(b).is_true();
}


//
// Fact Condition
//

int FactCondition::check(BindingList& bl) const
// compare the relation against world model and add/delete bindings
{
  BindingListIterator binding_iterator(bl);
  Binding* old_b;
  Binding* new_b;
  
  while (old_b = binding_iterator()) {
    
    WmTableBucketIterator next_wr(program.world_model, get_relation());

    for (new_b=new Binding(old_b); next_wr(new_b); new_b=new Binding(old_b))
      binding_iterator.insert_here(new_b);

    binding_iterator.remove_this();

    delete new_b;
    delete old_b;
  }
  
  return bl.get_count();
}

int FactCondition::confirm(Binding* b) const
// confirm whether the binding is still valid against the current WM
{
  return program.world_model->match(get_relation(), b);
}


//
// Retrieve Condition
//

int RetrieveCondition::check(BindingList& bl) const
// compare the relation against world model and add/delete bindings
{
  BindingListIterator binding_iterator(bl);
  Binding* old_b;
  Binding* new_b;
  
  while (old_b = binding_iterator()) {
    
    // unbind variables before matching
    old_b->unbind_variables(get_relation()->get_args());

    WmTableBucketIterator next_wr(program.world_model, get_relation());

    for (new_b=new Binding(old_b); next_wr(new_b); new_b=new Binding(old_b))
      binding_iterator.insert_here(new_b);

    binding_iterator.remove_this();

    delete new_b;
    delete old_b;
  }
  
  return bl.get_count();
}

int RetrieveCondition::confirm(Binding* b) const
// confirm whether the binding is still valid against the current WM
{
  // unbound variables before matching
  b->unbind_variables(get_relation()->get_args());

  // now try match to get new values
  return program.world_model->match(get_relation(), b);
}
