//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: condition.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/condition.h,v $
//  
//  File	      : condition.h
//  Author(s)	      : Jaeho Lee <jaeho@eecs.umich.edu>
//  
//  Description	      : 
//  
//  Original author(s): Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Wed Feb	 9 22:24:19 1994
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:34 1996
//  Update Count      : 102
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __CONDITION_H__
#define __CONDITION_H__

#include "dlist.h"
#include "binding.h"
#include "exp.h"
#include "wm.h"

//

class Condition
{
  Condition* 	rep;
  long 		active_value;

public:

  enum ConditionType
  {
    COND_GOAL,
    COND_EXP,
    COND_FACT,
    COND_RETRIEVE
  };

  virtual ~Condition()
  {}

  Condition()
  {
    rep = 0;
    active_value = 1;
  }

//

  Condition* set_positive()
  {
    active_value = 1;
    return this;
  }

  Condition* set_negative()
  {
    active_value = 0;
    return this;
  }
  
  virtual const char* get_name() const
  {
    return rep->get_name();
  }

  virtual const ConditionType type() const
  {
    return rep->type();
  }

  virtual int check(BindingList& bl) const
  {
    return rep->check(bl);
  }

  virtual int confirm(Binding* b) const
  {
    return rep->confirm(b);
  }
};

typedef Dlist<Condition*>		ConditionList;
typedef DlistIterator<Condition*>	ConditionListIterator;


//
// Expression Condition
//

class ExpCondition : public Condition
{
  Expression* expression;

public:

  ~ExpCondition()
  {
    delete expression;
  }
  
  ExpCondition(Expression* e)
  : Condition()
  {
    expression = e;
  }

  const	char* get_name() const
  {
    return expression->get_name();
  }
  
  const	ConditionType type() const
  {
    return COND_EXP;
  }

  int check(BindingList& bl) const;
  int confirm(Binding* b) const;
};  


//
// RelCondition
//

class RelCondition : public Condition
{
  Relation* relation;

public:

  virtual ~RelCondition()
  {
    delete relation;
  }
  
  RelCondition(Relation* r)
  : Condition()
  {
    relation = r;
  }

  const char* get_name() const
  {
    return relation->get_name();
  }

  Relation* get_relation() const
  {
    return relation;
  }

  virtual const ConditionType type() const = 0;

  virtual int check(BindingList& bl) const = 0;
  virtual int confirm(Binding* b) const = 0;
};

//

class FactCondition : public RelCondition
{
public:

  FactCondition(Relation* r)
  : RelCondition(r)
  {}

  const ConditionType type() const
  {
    return COND_FACT;
  }

  int check(BindingList& bl) const;
  int confirm(Binding* b) const;
};

//

class RetrieveCondition : public RelCondition
{
public:

  RetrieveCondition(Relation* r)
  : RelCondition(r)
  {}

  const ConditionType type() const
  {
    return COND_RETRIEVE;
  }

  int check(BindingList& bl) const;
  int confirm(Binding* b) const;
};

#endif

