//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: dlist.h,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/dlist.h,v $
//  
//  File	      : dlist.h
//  Author(s)	      : Jaeho Lee <jaeho@wham.engin.umich.edu>
//  
//  Description	      : 
//  
//  Original author(s): Jaeho Lee <jaeho@wham.engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sun Nov 27 01:09:43 1994
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:47:52 1997
//  Update Count      : 387
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __DLIST_H__
#define __DLIST_H__

//

template <class T> class Dlink
{
friend class Dlist<T>;
friend class DlistIterator<T>;

protected:
  
  Dlink<T>*	_prev;
  Dlink<T>*	_next;
  T		_ent;

  Dlink()
  {
    _prev = _next = this;
  }

  Dlink(T ent, Dlink<T>* prev, Dlink<T>* next)
  {
    _ent = ent;
    _prev = prev;
    _next = next;
  }
};


template <class T> class Dlist
// Doubly Linked List
{
friend class DlistIterator<T>;

protected:

  Dlink<T>*	_head;
  long		_count;

  //
  
  void create_head()	{ _head = new Dlink<T>; _count = 0; }

  void	    add_after	(Dlink<T>* current, T ent)
  {
    Dlink<T>* link = new Dlink<T>(ent, current, current->_next);

    current->_next->_prev = link;
    current->_next = link;

    _count++;
  }

  void	    add_after	(Dlink<T>* current, Dlist<T>& list)
  {
    DlistIterator<T> list_iterator(list);

    while(!list_iterator.end_of_list()) {
      add_after(current, list_iterator());
      current = current->_next;
    }
  }

  void	    add_before	(Dlink<T>* current, T ent)
  {
    Dlink<T>* link = new Dlink<T>(ent, current->_prev, current);
  
    current->_prev->_next = link;
    current->_prev = link;

    _count++;
  }

  void	    add_before	(Dlink<T>* current, Dlist<T>& list)
  {
    DlistIterator<T> list_iterator(list);

    while(!list_iterator.end_of_list())
      add_before(current, list_iterator());
  }

  Dlink<T>* remove	(Dlink<T>* current)
  {
    if (current == _head)
      return current;

    Dlink<T>* previous = current->_prev;
  
    previous->_next = current->_next;
    current->_next->_prev = current->_prev;

    delete current;
    _count--;

    return previous;
  }

  Dlink<T>* replace	(Dlink<T>* current, T ent)
  {
    if (current == _head)
      return current;
  
    Dlink<T>* link = new Dlink<T>(ent, current->_prev, current->_next);
    current->_prev->_next = link;
    current->_next->_prev = link;

    delete current;
    return current = link;
  }  

  
public:

  // add at head of list
  virtual void insert(T ent)		{ add_after(_head, ent); }
  virtual void insert(Dlist<T>& list)	{ add_after(_head, list); }

  // add at tail of list
  virtual void append(T ent)		{ add_before(_head, ent); }
  virtual void append(Dlist<T>& list)	{ add_before(_head, list); }
  //

  void	clear()
  // remove all links
  {
    Dlink<T>* current = _head->_next;

    while (current != _head) {
      Dlink<T>* n = current->_next;
      delete current;
      current = n;
    }

    _count = 0;
    _head->_prev = _head->_next = _head;
  }

  void	delete_entities()
  // delete all the entities in the list
  {
    Dlink<T>* current = _head->_next;

    while (current != _head) {
      delete current->_ent;
      current = current->_next;
    }
  }

  T	first(T eol_value=0)
  // return the first element of the list
  {
    if (is_empty())
      return eol_value;
    else
      return _head->_next->_ent;
  }

  T	last(T eol_value=0)
  // return the last element of the list
  {
    if (is_empty())
      return eol_value;
    else
      return _head->_prev->_ent;
  }

  T	pop(T eol_value=0)
  // return and remove head of list
  {
    if (is_empty())
      return eol_value;

    Dlink<T>* first = _head->_next;

    _head->_next = first->_next;
    first->_next->_prev = _head;

    T ent = first->_ent;
    delete first;
    _count--;

    return ent;
  }

  //
  
  long	get_count() const	{ return _count; }
  int	is_empty() const	{ return _count <= 0; }

  //
  //
  //
  
  virtual ~Dlist()		{ clear(); delete _head; }

  Dlist()			{ create_head(); }
  Dlist(const T& ent)		{ create_head(); insert(ent); }
};


///
///
/// Dlist Iterator
///
///

template <class T> class DlistIterator
{
  Dlist<T>*	_list;
  Dlink<T>*	_current;

protected:

  typedef int (*MatchFunction)(T, T);

public:

  // reinitialize the iterator
  void	reset()			{ _current = _list->_head; }
  void	reset(Dlist<T>& list)	{ _current = (_list=&list)->_head; }
  void	reset(Dlist<T>* list)	{ _current = (_list= list)->_head; }

  // True if there is no more element in the list
  int	end_of_list()			{ return _current->_next == _list->_head; }

  // romove the current link from the list
  void	remove_this()			{ _current = _list->remove(_current); }

  // replace the current one with an element
  void	replace_this(T ent)		{ _current = _list->replace(_current, ent); }

  // add an element before the current one
  void	insert_here(T ent)		{ _list->add_before(_current, ent); }
  void	insert_here(Dlist<T>& list)	{ _list->add_before(_current, list); }

  // add an element after the current one
  void	append_here(T ent)		{ _list->add_after(_current, ent); }
  void	append_here(Dlist<T>& list)	{ _list->add_after(_current, list); }
  
  T	get_this(T eol_value=0) 
  // return the current element in the list in sequence
  {
    if (_current == _list->_head)
      return eol_value;
    return _current->_ent;
  }

  T	get_next(T eol_value=0) 
  // return an element in the list in sequence
  {
    if (end_of_list())
      return eol_value;
    return (_current=_current->_next)->_ent;
  }

  T	get_next(MatchFunction match_function, T argument, T eol_value=0)
  // return an element in the list in sequence satisfying the matching function
  {
    while (!end_of_list()) {
      _current = _current->_next;
      if (match_function(_current->_ent, argument))
	return _current->_ent;
    }
    return eol_value;
  }

  T	operator()(T eol_value=0)
  {
    return get_next(eol_value);
  }
  
  T	operator()(MatchFunction match_function, T argument, T eol_value=0)
  {
    return get_next(match_function, argument, eol_value);
  }

  //
  //
  //

  DlistIterator(Dlist<T>& list) { reset(list); }
  DlistIterator(Dlist<T>* list) { reset(list); }
};


//
//
//
// Ordered Dlist
//
//
//


template <class T> class ODlist : public Dlist<T>
// Ordered Doubly Linked List
{
friend class DlistIterator<T>;

public:
  
  typedef int& (*OrdinalFunction)(T);

protected:
  
  OrdinalFunction	_ordinal_function;

public:

  void	set_ordinal_function(OrdinalFunction ordinal_function)
  {
    _ordinal_function = ordinal_function;
  }
  
  void	insert(T ent)
  {
    if (!_ordinal_function) {
      Dlist<T>::insert(ent);
      return;
    }
  
    DlistIterator<T> list_iterator(this);
    int ordinal = _ordinal_function(ent);

    while(!list_iterator.end_of_list()) {
      if (ordinal < _ordinal_function(list_iterator())) {
	list_iterator.insert_here(ent);
	return;
      }
    }
    list_iterator.append_here(ent);
  }

  void	insert(Dlist<T>& list)
  {
    DlistIterator<T> list_iterator(list);

    while(!list_iterator.end_of_list())
      insert(list_iterator());
  }
    
  void	renumber(long start=0, long inc=1)
  {
    if (!_ordinal_function)
      return;

    DlistIterator<T> list_iterator(this);

    while(!list_iterator.end_of_list()) {
      _ordinal_function(list_iterator()) = start;
      start += inc;
    }
  }

  //
  //
  //

  ODlist(OrdinalFunction ordinal_function=0)
  : Dlist<T>()	     
  {
    _ordinal_function = ordinal_function;
  }
  
  ODlist(T& ent, OrdinalFunction ordinal_function=0)
  : Dlist<T>(ent)
  {
    _ordinal_function = ordinal_function;
  }
};



#endif
