//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: exp-fun.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/exp-fun.h,v $
//  
//  File	      : exp-fun.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : Expression (Function Call)
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:33:23 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:24 1996
//  Update Count      : 78
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __EXP_FUN_H__
#define __EXP_FUN_H__

#include <stream.h>
#include <string.h>
#include "exp.h"
#include "exp-val.h"

/*

 funcall:
	(funcation expression-list)

*/

// Classes and types

class Value;
class Variable;
class Funcall;

class Binding;

//
typedef Value (*Function)(long arity, ExpList* args, Binding* binding);


//
// FUNCTION CALL
//

class Funcall : public Expression
{
  char*		_name;
  long		_arity;
  Function	_function;
  ExpList*	_args;

  void init(char* name, long arity=0, Function function=0, ExpList* args=0)
  {
    ::strcpy(_name=new char[::strlen(name)+1], name);
    _arity = arity;
    _function = function;
    _args = args;
  }

public:

  virtual ~Funcall()
  {
    delete _name;
    _args->delete_entities();
    delete _args;
  }
  
  Funcall(char*	name)
  {
    init(name);
  }

  Funcall(char*	name, Function function, ExpList* args)
  {
    init(name, args?args->get_count():0, function, args);
  }

  Funcall(Funcall& function)
  {
    init(function._name, function._arity, function._function, function._args);
  }

  void operator=(Funcall& function)
  {
    if (this == &function) return;
    delete _name;
    init(function._name, function._arity, function._function, function._args);
  }
  
  // access functions
  const char* get_name() const
  {
    return _name;
  }
  
  long get_arity() const
  {
    return _arity;
  }

  const ExpType get_type() const
  {
    return EXP_FUNCALL;
  }
  
  Value eval(Binding*const binding=0)
  {
    return _function(_arity, _args, binding);
  }
  
  const int is_variable() const
  {
    return 0;
  }
  

  void print(ostream& s=cout, Binding* b=0)
  {
    eval(b).print(s, b);
  }

  void format(ostream& s=cout, Binding* b=0)
  {
    eval(b).format(s, b);
  }
  
friend ostream& operator<<(ostream& s, Funcall& f);
};

#endif
