//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//
//  $Id: exp-val.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/exp-val.cc,v $
//
//  File	      : exp-val.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//
//  Description	      : Expression (Value)
//
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:33:00 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:05:31 1996
//  Update Count      : 378
//
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
//
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>
#include <stdio.h>
#include <ctype.h>

#include "exp-val.h"


//
// Static Member
// 
Value const Value::Undefined	= Value();
Value const Value::True		= Value(1);
Value const Value::False	= Value(0);

//
// Void Value
//

class VoidValue: public Value
{
friend class IntValue;
friend class RealValue;
friend class StrValue;

private:

  void* _val;

public:

  VoidValue(void* v)
  : Value(BaseValue())
  {
    _val = v;
  }

public:
  
  void	print(ostream& s=cout, Binding* b=0)
  {
    int old_options = s.flags(ios::hex|ios::showbase);
    s << _val;
    s.flags(old_options);
  }

  void format(ostream& s=cout, Binding* b=0)
  {
    print(s, b);
  }

  const ValType type()		const { return VAL_VOID;	}
  const int	is_true()	const { return _val != 0;	}

public:
  
  void*	 get_void  ()		    const { return _val;		}
  long	 get_int   ()		    const { return 0;			}
  double get_real  ()		    const { return 0.0;			}
  char*	 get_string()		    const { return 0;			}
		     
  int	 operator!()		    const { return !is_true();		}

  int	 operator==(const Value& v) const { return v.voidEq(*this);	}
  int	 operator!=(const Value& v) const { return v.voidNe(*this);	}

public:	 

  int	 voidEq	   (const Value& v) const;
  int	 voidNe	   (const Value& v) const;
};


//
// Integer Value
//

class IntValue: public Value
{
friend class VoidValue;
friend class RealValue;
friend class StrValue;

private:

  long _val;

public:

  IntValue(const IntValue& v)
  : Value(BaseValue())
  {
    _val = v._val;
  }

  IntValue(long v)
  : Value(BaseValue())
  {
    _val = v;
  }

public:
  
  void print(ostream& s=cout, Binding* b=0)
  {
    s << _val;
  }

  void format(ostream& s=cout, Binding* b=0)
  {
    print(s, b);
  }

  const ValType type()		    const { return VAL_INT;		}
  const int	is_true()	    const { return _val != 0;		}

public:
	
  void*	 get_void  ()		    const { return 0;			}
  long	 get_int   ()		    const { return _val;		}
  double get_real  ()		    const { return (double) _val;	}
  char*	 get_string()		    const { return 0;			}
	 
  Value	 operator- ()		    const { return Value(-_val);	}
  Value	 operator+ (const Value& v) const { return v.intAdd(*this);	}
  Value	 operator- (const Value& v) const { return v.intSub(*this);	}
  Value	 operator* (const Value& v) const { return v.intMul(*this);	}
  Value	 operator/ (const Value& v) const { return v.intDiv(*this);	}
  Value	 operator% (const Value& v) const { return v.intMod(*this);	}
				    
  int	 operator! ()		    const { return !is_true();		}
  int	 operator< (const Value& v) const { return v.intLt(*this);	}
  int	 operator> (const Value& v) const { return v.intGt(*this);	}
  int	 operator<=(const Value& v) const { return v.intLe(*this);	}
  int	 operator>=(const Value& v) const { return v.intGe(*this);	}
  int	 operator==(const Value& v) const { return v.intEq(*this);	}
  int	 operator!=(const Value& v) const { return v.intNe(*this);	}
	 
public:	 
	 
  Value	 intAdd	   (const Value& v) const;
  Value	 intSub	   (const Value& v) const;
  Value	 intMul	   (const Value& v) const;
  Value	 intDiv	   (const Value& v) const;
  Value	 intMod	   (const Value& v) const;
				       
  Value	 realAdd   (const Value& v) const;
  Value	 realSub   (const Value& v) const;
  Value	 realMul   (const Value& v) const;
  Value	 realDiv   (const Value& v) const;
				       
  Value	 strAdd	   (const Value& v) const;
				       
  int    voidEq	   (const Value& v) const;
  int    voidNe	   (const Value& v) const;

  int	 intLt	   (const Value& v) const;
  int	 intGt	   (const Value& v) const;
  int	 intLe	   (const Value& v) const;
  int	 intGe	   (const Value& v) const;
  int	 intEq	   (const Value& v) const;
  int	 intNe	   (const Value& v) const;
				       
  int	 realLt	   (const Value& v) const;
  int	 realGt	   (const Value& v) const;
  int	 realLe	   (const Value& v) const;
  int	 realGe	   (const Value& v) const;
  int	 realEq	   (const Value& v) const;
  int	 realNe	   (const Value& v) const;
				     
  int	 strEq	   (const Value& v) const;
  int	 strNe	   (const Value& v) const;
};


//
// Real Value
//

class RealValue: public Value
{
friend class VoidValue;
friend class IntValue;
friend class StrValue;
  
private:

  double _val;

public:

  RealValue(const RealValue& v)
  : Value(BaseValue())
  {
    _val = v._val;
  }

  RealValue(double v)
  : Value(BaseValue())
  {
    _val = v;
  }

public:
  
  void print(ostream& s=cout, Binding* b=0)
  {
    s << _val;
  }

  void format(ostream& s=cout, Binding* b=0)
  {
    print(s, b);
  }

  const ValType type()		    const { return VAL_REAL;		}
  const int	is_true()	    const { return _val != 0.0;		}

public:
  
  void*	 get_void  ()		    const { return 0;			}
  long	 get_int   ()		    const { return (long) _val;		}
  double get_real  ()		    const { return _val;		}
  char*	 get_string()		    const { return 0;			}
	 
  Value	 operator- ()		    const { return Value(-_val);	}
  Value	 operator+ (const Value& v) const { return v.realAdd(*this);	}
  Value	 operator- (const Value& v) const { return v.realSub(*this);	}
  Value	 operator* (const Value& v) const { return v.realMul(*this);	}
  Value	 operator/ (const Value& v) const { return v.realDiv(*this);	}	
				   
  int	 operator! ()		    const { return !is_true();		}
  int	 operator< (const Value& v) const { return v.realLt(*this);	}
  int	 operator> (const Value& v) const { return v.realGt(*this);	}
  int	 operator<=(const Value& v) const { return v.realLe(*this);	}
  int	 operator>=(const Value& v) const { return v.realGe(*this);	}
  int	 operator==(const Value& v) const { return v.realEq(*this);	}
  int	 operator!=(const Value& v) const { return v.realNe(*this);	}	
	 
public:	 
	 
  Value	 intAdd	   (const Value& v) const;
  Value	 intSub	   (const Value& v) const;
  Value	 intMul	   (const Value& v) const;
  Value	 intDiv	   (const Value& v) const;
				     
  Value	 realAdd   (const Value& v) const;
  Value	 realSub   (const Value& v) const;
  Value	 realMul   (const Value& v) const;
  Value	 realDiv   (const Value& v) const;
				     
  Value	 strAdd	   (const Value& v) const;
				     
  int    voidEq	   (const Value& v) const;
  int    voidNe	   (const Value& v) const;

  int	 intLt	   (const Value& v) const;
  int	 intGt	   (const Value& v) const;
  int	 intLe	   (const Value& v) const;
  int	 intGe	   (const Value& v) const;
  int	 intEq	   (const Value& v) const;
  int	 intNe	   (const Value& v) const;
				     
  int	 realLt	   (const Value& v) const;
  int	 realGt	   (const Value& v) const;
  int	 realLe	   (const Value& v) const;
  int	 realGe	   (const Value& v) const;
  int	 realEq	   (const Value& v) const;
  int	 realNe	   (const Value& v) const;
				     
  int	 strEq	   (const Value& v) const;
  int	 strNe	   (const Value& v) const;
};


//
// String Value
//

class StrValue: public Value
{
friend class VoidValue;
friend class IntValue;
friend class RealValue;

private:

  char* _val;

private:
  
  void concat_strings(char* a, char* b)
  {
    ::strcpy(_val=new char[::strlen(a)+::strlen(b)+1], a);
    ::strcpy(_val+::strlen(a), b);
  }

  // temporary buffer to do string operations
  static char* strValueBuffer;

public:

  ~StrValue()
  {
    delete _val;
  }

  StrValue(const StrValue& v)
  : Value(BaseValue())
  {
    ::strcpy(_val=new char[::strlen(v._val)+1], v._val);
  }

  StrValue(char* v)
  : Value(BaseValue())
  {
    ::strcpy(_val=new char[::strlen(v)+1], v);
  }

  StrValue(char* a, char* b)
  : Value(BaseValue())
  {
    concat_strings(a, b);
  }

  StrValue(char* a, long int_b)
  : Value(BaseValue())
  {
    sprintf(strValueBuffer, "%ld", int_b);
    concat_strings(a, strValueBuffer);
  }

  StrValue(char* a, double real_b)
  : Value(BaseValue())
  {
    sprintf(strValueBuffer, "%lf", real_b);
    concat_strings(a, strValueBuffer);
  }

  StrValue(long int_a, char* b)
  : Value(BaseValue())
  {
    sprintf(strValueBuffer, "%ld", int_a);
    concat_strings(strValueBuffer, b);
  }

  StrValue(double real_a, char* b)
  : Value(BaseValue())
  {
    sprintf(strValueBuffer, "%lf", real_a);
    concat_strings(strValueBuffer, b);
  }

public:
  
  void print(ostream& s=cout, Binding* b=0)
  {
    s << _val;
  }

  virtual void format(ostream& s=cout, Binding* b=0)
  {
    s << '"';
    for (char* p = _val; *p; p++) {
      if (*p == '"')
	s << '\\' << '"';
      else if (isascii(*p) && isprint(*p))
	s << *p;
      else {
	s << '\\' << 'x';
	s << hex((unsigned char)*p);
      }
    }
    s << '"';
  }

  const ValType type()		    const { return VAL_STRING;		}
  const int	is_true()	    const { return ::strlen(_val) != 0;	}
	
public:
  
  void*	 get_void  ()		    const { return (void*) _val;	}
  long	 get_int   ()		    const { return 0;			}
  double get_real  ()		    const { return 0.0;			}
  char*	 get_string()		    const { return _val;		}
  	 
  Value	 operator+ (const Value& v) const { return v.strAdd(*this);	}
	 				 
  int	 operator! ()		    const { return !is_true();		}
  int	 operator< (const Value& v) const { return v.strLt(*this);	}
  int	 operator> (const Value& v) const { return v.strGt(*this);	}
  int	 operator<=(const Value& v) const { return v.strLe(*this);	}
  int	 operator>=(const Value& v) const { return v.strGe(*this);	}
  int	 operator==(const Value& v) const { return v.strEq(*this);	}
  int	 operator!=(const Value& v) const { return v.strNe(*this);	}	
					   
public:					   
					   
  Value	 intAdd	   (const Value& v) const;
  Value	 realAdd   (const Value& v) const;
  Value	 strAdd	   (const Value& v) const;
					   
  int    voidEq	   (const Value& v) const;
  int    voidNe	   (const Value& v) const;

  int	 intEq	   (const Value& v) const;
  int	 intNe	   (const Value& v) const;
				     
  int	 realEq	   (const Value& v) const;
  int	 realNe	   (const Value& v) const;

  int	 strLt	   (const Value& v) const;
  int	 strGt	   (const Value& v) const;
  int	 strLe	   (const Value& v) const;
  int	 strGe	   (const Value& v) const;
  int	 strEq	   (const Value& v) const;
  int	 strNe	   (const Value& v) const;
};

//

char* StrValue::strValueBuffer = new char[100];


//////////////////////////////////////////////////////////////////////
/////								 /////
/////		Member Functions				 /////
/////								 /////
//////////////////////////////////////////////////////////////////////

Value::Value()			{ _rep = new VoidValue(0); }	// void					
Value::Value(void*  v)		{ _rep = new VoidValue(v); }	// void					
Value::Value(int    i)		{ _rep = new IntValue (i); }	// integer
Value::Value(long   i)		{ _rep = new IntValue (i); }	// integer				
Value::Value(double f)		{ _rep = new RealValue(f); }	// real					
Value::Value(char*  s)		{ _rep = new StrValue (s); }	// string				

//
// Void Value
//

int   VoidValue::voidEq	(const Value& v) const { return ((VoidValue*) &v)->_val == _val; }
int   VoidValue::voidNe	(const Value& v) const { return ((VoidValue*) &v)->_val != _val; }

//
// Integer Value
//

/*
Value IntValue::intAdd(const Value& v) const
{
  Value retval;
  IntValue* v1 = new IntValue(*this);
  IntValue* v2 = (IntValue*)&v;
  v1->_val += v2->_val;
  retval.redefine(v1);
  return retval;
}

Value IntValue::intSub(const Value& v) const
{
  Value retval;
  IntValue* v1 = new IntValue(*((IntValue*)&v));
  v1->_val -= _val;
  retval.redefine(v1);
  return retval;
}
*/

Value IntValue::intAdd	(const Value& v) const { return Value(((IntValue*) &v)->_val + _val);	}
Value IntValue::intSub	(const Value& v) const { return Value(((IntValue*) &v)->_val - _val);	}
Value IntValue::intMul	(const Value& v) const { return Value(((IntValue*) &v)->_val * _val);	}
Value IntValue::intDiv	(const Value& v) const { return Value(((IntValue*) &v)->_val / _val);	}
Value IntValue::intMod	(const Value& v) const { return Value(((IntValue*) &v)->_val % _val);	}
					       
Value IntValue::realAdd (const Value& v) const { return Value(((RealValue*)&v)->_val + _val);	}
Value IntValue::realSub (const Value& v) const { return Value(((RealValue*)&v)->_val - _val);	}
Value IntValue::realMul (const Value& v) const { return Value(((RealValue*)&v)->_val * _val);	}
Value IntValue::realDiv (const Value& v) const { return Value(((RealValue*)&v)->_val / _val);	}	
					     
Value IntValue::strAdd(const Value& v) const
{
  Value retval;
  retval.redefine(new StrValue(((StrValue*) &v)->_val, _val));
  return retval;
}
					     
int   IntValue::voidEq	(const Value& v) const { return (long) ((VoidValue*) &v)->_val == _val; }
int   IntValue::voidNe	(const Value& v) const { return (long) ((VoidValue*) &v)->_val != _val; }

int   IntValue::intLt	(const Value& v) const { return ((IntValue*) &v)->_val <  _val;		}
int   IntValue::intGt	(const Value& v) const { return ((IntValue*) &v)->_val >  _val;		}
int   IntValue::intLe	(const Value& v) const { return ((IntValue*) &v)->_val <= _val;		}
int   IntValue::intGe	(const Value& v) const { return ((IntValue*) &v)->_val >= _val;		}
int   IntValue::intEq	(const Value& v) const { return ((IntValue*) &v)->_val == _val;		}
int   IntValue::intNe	(const Value& v) const { return ((IntValue*) &v)->_val != _val;		}
						     
int   IntValue::realLt	(const Value& v) const { return ((RealValue*)&v)->_val <  _val;		}
int   IntValue::realGt	(const Value& v) const { return ((RealValue*)&v)->_val >  _val;		}
int   IntValue::realLe	(const Value& v) const { return ((RealValue*)&v)->_val <= _val;		}
int   IntValue::realGe	(const Value& v) const { return ((RealValue*)&v)->_val >= _val;		}
int   IntValue::realEq	(const Value& v) const { return ((RealValue*)&v)->_val == _val;		}
int   IntValue::realNe	(const Value& v) const { return ((RealValue*)&v)->_val != _val;		}	
					     
int   IntValue::strEq	(const Value& v) const { return 0; }
int   IntValue::strNe	(const Value& v) const { return 1; }
					    
//					    
// Real Value				    
//					    
Value RealValue::intAdd (const Value& v) const { return Value(((IntValue*) &v)->_val + _val);	}
Value RealValue::intSub (const Value& v) const { return Value(((IntValue*) &v)->_val - _val);	}
Value RealValue::intMul (const Value& v) const { return Value(((IntValue*) &v)->_val * _val);	}
Value RealValue::intDiv (const Value& v) const { return Value(((IntValue*) &v)->_val / _val);	}
					      
Value RealValue::realAdd(const Value& v) const { return Value(((RealValue*)&v)->_val + _val);	}
Value RealValue::realSub(const Value& v) const { return Value(((RealValue*)&v)->_val - _val);	}
Value RealValue::realMul(const Value& v) const { return Value(((RealValue*)&v)->_val * _val);	}
Value RealValue::realDiv(const Value& v) const { return Value(((RealValue*)&v)->_val / _val);	}	

Value RealValue::strAdd(const Value& v) const
{
  Value retval;
  retval.redefine(new StrValue(((StrValue*) &v)->_val, _val));
  return retval;
}
					    
int   RealValue::voidEq	(const Value& v) const { return (long) ((VoidValue*) &v)->_val == _val; }
int   RealValue::voidNe	(const Value& v) const { return (long) ((VoidValue*) &v)->_val != _val; }

int   RealValue::intLt	(const Value& v) const { return ((IntValue*) &v)->_val <  _val;		}
int   RealValue::intGt	(const Value& v) const { return ((IntValue*) &v)->_val >  _val;		}
int   RealValue::intLe	(const Value& v) const { return ((IntValue*) &v)->_val <= _val;		}
int   RealValue::intGe	(const Value& v) const { return ((IntValue*) &v)->_val >= _val;		}
int   RealValue::intEq	(const Value& v) const { return ((IntValue*) &v)->_val == _val;		}
int   RealValue::intNe	(const Value& v) const { return ((IntValue*) &v)->_val != _val;		}
						    
int   RealValue::realLt (const Value& v) const { return ((RealValue*)&v)->_val <  _val;		}
int   RealValue::realGt (const Value& v) const { return ((RealValue*)&v)->_val >  _val;		}
int   RealValue::realLe (const Value& v) const { return ((RealValue*)&v)->_val <= _val;		}
int   RealValue::realGe (const Value& v) const { return ((RealValue*)&v)->_val >= _val;		}
int   RealValue::realEq (const Value& v) const { return ((RealValue*)&v)->_val == _val;		}
int   RealValue::realNe (const Value& v) const { return ((RealValue*)&v)->_val != _val;		}	
					    
int   RealValue::strEq	(const Value& v) const { return 0; }
int   RealValue::strNe	(const Value& v) const { return 1; }
					    
//					    
// String Value				    
//
Value StrValue::intAdd(const Value& v) const
{
  Value retval;
  retval.redefine(new StrValue(((IntValue*) &v)->_val, _val));
  return retval;
}

Value StrValue::realAdd(const Value& v) const
{
  Value retval;
  retval.redefine(new StrValue(((RealValue*) &v)->_val, _val));
  return retval;
}

Value StrValue::strAdd(const Value& v) const
{
  Value retval;
  retval.redefine(new StrValue(((StrValue*) &v)->_val, _val));
  return retval;
}


int   StrValue::voidEq	(const Value& v) const { return 0; }
int   StrValue::voidNe	(const Value& v) const { return 1; }

int   StrValue::intEq	(const Value& v) const { return 0; }
int   StrValue::intNe	(const Value& v) const { return 1; }

int   StrValue::realEq	(const Value& v) const { return 0; }
int   StrValue::realNe	(const Value& v) const { return 1; }

int   StrValue::strLt	(const Value& v) const { return strcmp(((StrValue*)&v)->_val, _val) <  0; }
int   StrValue::strGt	(const Value& v) const { return strcmp(((StrValue*)&v)->_val, _val) >  0; }
int   StrValue::strLe	(const Value& v) const { return strcmp(((StrValue*)&v)->_val, _val) <= 0; }
int   StrValue::strGe	(const Value& v) const { return strcmp(((StrValue*)&v)->_val, _val) >= 0; }
int   StrValue::strEq	(const Value& v) const { return strcmp(((StrValue*)&v)->_val, _val) == 0; }
int   StrValue::strNe	(const Value& v) const { return strcmp(((StrValue*)&v)->_val, _val) != 0; }

//
//
//

ostream& operator<<(ostream& s, Value v)
{
  v.print(s);
  return s;
}
