//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: exp-val.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/exp-val.h,v $
//  
//  File	      : exp-val.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : Expression (Value)
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:33:00 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:21 1996
//  Update Count      : 208
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __EXP_VAL_H__
#define __EXP_VAL_H__

#include <string.h>
#include "exp.h"

class Binding;

//

struct BaseValue
{
  BaseValue(int=0) {}
};

//

class Value : public Expression
{
private:

  Value*	_rep;
  long		_referenceCount;

public:

  static const Value Undefined;
  static const Value True;
  static const Value False;

  enum ValType
  {
    VAL_VOID,
    VAL_INT,
    VAL_REAL,
    VAL_STRING,
    VAL_UNDEF
  };

  virtual long deref()
  {
    return --_referenceCount;
  }
  
  virtual long ref()
  {
    return ++_referenceCount;
  }
  
public:

  virtual ~Value()
  {
    if (_rep && _rep->deref() <= 0)
      delete _rep;
  }

  Value(BaseValue)
  : _rep(0), _referenceCount(1)
  {}
  
  Value();			// void
  Value(void*  v);		// void
  Value(int    i);		// integer
  Value(long   i);		// integer
  Value(double f);		// real
  Value(char* s);		// string

  //
  
  Value& operator=(const Value& x)
  {
    if (_rep != x._rep) {
      if (_rep && _rep->deref() <= 0)
	delete _rep;
      (_rep = x._rep)->ref();
    }
    return *this;
  }

  Value(const Value& x)
  {
    (_rep = x._rep)->ref();
  }

  void redefine(Value* x)
  {
    if (_rep && _rep->deref() <= 0)
      delete _rep;
    _rep = x;
  }

  Value eval(Binding*const b=0)
  {
    return *this;
  }

  const	  char*		get_name()	const { return "Value";			}
  const	  ExpType	get_type()	const { return EXP_VALUE;		}

  virtual const ValType	type()		const { return _rep->type();		}
  virtual const int	is_true()	const { return _rep->is_true();		}
  virtual const int	is_variable()	const { return 0;			}

  virtual const int	is_defined()	const
  {
    return !(_rep->type() == VAL_VOID && _rep->get_void() == 0);
  }

public:

  virtual void*	 get_void  ()	    	    const { return _rep->get_void();	}
  virtual long	 get_int   ()	    	    const { return _rep->get_int();	}
  virtual double get_real  ()	            const { return _rep->get_real();	}
  virtual char*	 get_string()	            const { return _rep->get_string();	}
		   	   
  virtual operator void*   ()		    const { return _rep->get_void();	}
  virtual operator int	   ()		    const { return _rep->get_int();	}
  virtual operator float   ()		    const { return _rep->get_real();	}
  virtual operator double  ()	    	    const { return _rep->get_real();	}
  virtual operator char*   ()		    const { return _rep->get_string();	}

public:

  virtual Value	operator-  ()		    const { return _rep->operator-();	}
  virtual Value	operator+  (const Value& v) const { return _rep->operator+(v);	}
  virtual Value	operator-  (const Value& v) const { return _rep->operator-(v);	}
  virtual Value	operator*  (const Value& v) const { return _rep->operator*(v);	}
  virtual Value	operator/  (const Value& v) const { return _rep->operator/(v);	}
  virtual Value	operator%  (const Value& v) const { return _rep->operator%(v);	}
			   		    
  virtual int	operator!  ()		    const { return _rep->operator! ();	}
  virtual int	operator== (const Value& v) const { return _rep->operator==(v);	}
  virtual int	operator!= (const Value& v) const { return _rep->operator!=(v);	}
  virtual int	operator<  (const Value& v) const { return _rep->operator< (v);	}
  virtual int	operator<= (const Value& v) const { return _rep->operator<=(v);	}
  virtual int	operator>  (const Value& v) const { return _rep->operator> (v);	}
  virtual int	operator>= (const Value& v) const { return _rep->operator>=(v);	}

  //
  
  friend ostream& operator<<(ostream& s, Value v);

  virtual void print(ostream& s=cout, Binding* b=0)
  {
    _rep->print(s, b);
  }

  virtual void format(ostream& s=cout, Binding* b=0)
  {
    _rep->format(s, b);
  }

public:
  
  virtual Value	intAdd	   (const Value& v) const { return _rep->intAdd(v);	}
  virtual Value	intSub	   (const Value& v) const { return _rep->intSub(v);	}
  virtual Value	intMul	   (const Value& v) const { return _rep->intMul(v);	}
  virtual Value	intDiv	   (const Value& v) const { return _rep->intDiv(v);	}
  virtual Value	intMod	   (const Value& v) const { return _rep->intMod(v);	}
			   		    
  virtual Value	realAdd	   (const Value& v) const { return _rep->realAdd(v);	}
  virtual Value	realSub	   (const Value& v) const { return _rep->realSub(v);	}
  virtual Value	realMul	   (const Value& v) const { return _rep->realMul(v);	}
  virtual Value	realDiv	   (const Value& v) const { return _rep->realDiv(v);	}
  virtual Value	realMod	   (const Value& v) const { return _rep->realMod(v);	}
			   		    
  virtual Value	strAdd	   (const Value& v) const { return _rep->strAdd(v);	}
  virtual Value	strSub	   (const Value& v) const { return _rep->strSub(v);	}
  virtual Value	strMul	   (const Value& v) const { return _rep->strMul(v);	}
  virtual Value	strDiv	   (const Value& v) const { return _rep->strDiv(v);	}
  virtual Value	strMod	   (const Value& v) const { return _rep->strMod(v);	}
			  		    
  virtual int	voidEq	   (const Value& v) const { return _rep->voidEq(v);	}
  virtual int	voidNe	   (const Value& v) const { return _rep->voidNe(v);	}
			   		    					
  virtual int	intEq	   (const Value& v) const { return _rep->intEq(v);	}
  virtual int	intNe	   (const Value& v) const { return _rep->intNe(v);	}
  virtual int	intLt	   (const Value& v) const { return _rep->intLt(v);	}
  virtual int	intLe	   (const Value& v) const { return _rep->intLe(v);	}
  virtual int	intGt	   (const Value& v) const { return _rep->intGt(v);	}
  virtual int	intGe	   (const Value& v) const { return _rep->intGe(v);	}
			   		    					
  virtual int	realEq	   (const Value& v) const { return _rep->realEq(v);	}
  virtual int	realNe	   (const Value& v) const { return _rep->realNe(v);	}
  virtual int	realLt	   (const Value& v) const { return _rep->realLt(v);	}
  virtual int	realLe	   (const Value& v) const { return _rep->realLe(v);	}
  virtual int	realGt	   (const Value& v) const { return _rep->realGt(v);	}
  virtual int	realGe	   (const Value& v) const { return _rep->realGe(v);	}
			   		    					
  virtual int	strLt	   (const Value& v) const { return _rep->strLt(v);	}
  virtual int	strLe	   (const Value& v) const { return _rep->strLe(v);	}
  virtual int	strGt	   (const Value& v) const { return _rep->strGt(v);	}
  virtual int	strGe	   (const Value& v) const { return _rep->strGe(v);	}
  virtual int	strEq	   (const Value& v) const { return _rep->strEq(v);	}
  virtual int	strNe	   (const Value& v) const { return _rep->strNe(v);	}
};

#endif

