//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: exp-var.h,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/exp-var.h,v $
//  
//  File	      : exp-var.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : Expression (Variable)
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:32:22 1993
//  Last Modified By  : Jaeho Lee <jaeho@splat.engin.umich.edu>
//  Last Modified On  : Mon Feb 26 22:35:58 1996
//  Update Count      : 60
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __EXP_VAR_H__
#define __EXP_VAR_H__

#include <stream.h>
#include "binding.h"
#include "symtab.h"
#include "exp.h"
#include "exp-val.h"

//
// VARIABLE
//

class Variable : public Expression
{
  SymbolTable*	_st;
  long		_id;

  void init(SymbolTable* t, char* s)
  {
    _st = t;
    if ((_id=_st->get_id(s)) < 0)
      _id = _st->add(new Symbol(s))->get_id();
  }
  
public:

  Variable(SymbolTable* t, char* s)
  {
    init(t, s);
  }

  Variable(Binding* b, char* s)
  {
    init(b->get_symtab(), s);
  }

  const long get_id() const
  {
    return _id;
  }
  
  const char* get_name() const
  {
    return _st->lookup(get_id())->get_name();
  }
  
  const ExpType get_type() const
  {
    return EXP_VARIABLE;
  }
  
  Value eval(Binding*const b=0)
  {
    return b ? b->get_value(this) : Value::Undefined;
  }
  
  const int is_variable() const
  {
    return 1;
  }
  
  const Variable*const get_variable() const
  {
    return this;
  }

public:

  void print(ostream& s=cout, Binding* b=0)
  {
    eval(b).print(s, b);
  }
  
  void format(ostream& s=cout, Binding* b=0)
  {
    eval(b).format(s, b);
  }
  
friend ostream& operator<<(ostream& s, Variable& v);
};

#endif

