//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//
//  $Id: exp.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/exp.cc,v $
//
//  File	      : exp.cc
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//
//  Description	      : Expression
//
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Tue Oct 24 15:43:40 1995
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:05:28 1996
//  Update Count      : 17
//
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
//
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include "exp.h"
#include "exp-val.h"

//
// 	Expression List functions
//

ExpList* explist_eval(ExpList* explist, Binding*const binding)
{
  if (! explist) return 0;
  
  register Expression* e;
  ExpListIterator next_exp(explist);

  Value* v;
  while(e=next_exp()) {
    v = new Value(e->eval(binding));
    delete e;
    e = v;
  }

  return explist;
}

ExpList* explist_eval_new(ExpList* explist, Binding*const binding)
{
  if (! explist) return 0;
  
  register Expression* e;
  ExpListIterator next_exp(explist);
  ExpList* new_explist = new ExpList;

  while (e=next_exp())
    new_explist->append(new Value(e->eval(binding)));

  return new_explist;
}

void explist_delete(ExpList* explist)
{
  if (explist) {
    explist->delete_entities();
    delete explist;
  }
}

void explist_print(ExpList* explist, ostream& s, Binding* b)
{
  if (! explist) return;
  
  register Expression* e;
  ExpListIterator next_exp(explist);

  while(e=next_exp()) 
    s << e->eval(b) << " ";
}

void explist_format(ExpList* explist, ostream& s, Binding* b)
{
  if (! explist) return;
  
  register Expression* e;
  ExpListIterator next_exp(explist);

  while(e=next_exp()) {
    e->eval(b).format(s, b);
    s << " ";
  }
}
