//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: exp.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/exp.h,v $
//  
//  File	      : exp.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : Expression
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 10:33:00 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:15 1996
//  Update Count      : 74
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __EXP_H__
#define __EXP_H__

#include <stream.h>
#include <string.h>
#include "dlist.h"

/*
 expression:
	value
	variable
	funcall

 value:
	undefined
	char
	int
	float
	string

 funcall:
	(funcation expression-list)

 expression_list:
	NULL
	expression
	expression_list expression
*/

// Classes and types

class Value;
class Variable;
class Funcall;

class Binding;

//
// Expression
//

class Expression
{
public:

  virtual ~Expression()
  {};

  enum ExpType
  {
    EXP_VALUE,
    EXP_VARIABLE,
    EXP_FUNCALL,
    EXP_PREDICATE,
    EXP_UNDEFINED
  };

  virtual Value			eval(Binding*const b=0)  	= 0;

  virtual const char*		get_name() const		= 0;
  virtual const ExpType		get_type() const		= 0;

  // For variables only
  virtual const int		is_variable() const		= 0;

  virtual const Variable*const	get_variable() const	{ return 0; }

  virtual void			print(ostream& s=cout, Binding* b=0) = 0;
  virtual void			format(ostream& s=cout, Binding* b=0) = 0;
    
  friend  ostream&		operator<<(ostream&, Expression&);
};


//
//	Expression List (elist)
//

typedef Dlist<Expression*>		ExpList;
typedef DlistIterator<Expression*>	ExpListIterator;

//
// 	Expression List functions
//

ExpList* explist_eval(ExpList* explist, Binding*const binding);
ExpList* explist_eval_new(ExpList* explist, Binding*const binding);

void explist_delete(ExpList* explist);

void explist_print(ExpList* explist, ostream& s, Binding* b);
void explist_format(ExpList* explist, ostream& s, Binding* b);

#endif
