//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: goal.h,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/goal.h,v $
//  
//  File	      : goal.h
//  Author(s)	      : Marcus J. Huber <marcush@engin.umich.edu>
//			Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : Goal Structure
//  
//  Original author(s): Marcus J. Huber <marcush@engin.umich.edu>
//			Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:29:21 1993
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:50:10 1997
//  Update Count      : 259
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __GOAL_H__
#define __GOAL_H__

//

#include "dlist.h"
#include "action.h"

#include "is.h"
#include "soak.h"
#include "run.h"

// Forward declarations

class Relation;
class Binding;
class WmTable;

//
// Goal
//
class Goal
{

  // Change this to goal_action
  GoalAction*		goal;		// The expression for the goal to be
					// completed - points to action from parent
  					// KA body
  Goal*			subgoal;
  
  Goal*			prev_goal;	// Goal for which this intention was created

  int			new_goal;	// Flag of whether goal is new or not.

  StatusType		status;		// Run state of the intention (ACTIVE,
					// SUSPENDED, SUCCESS, FAILURE, etc.)

  SoakElement*		intention;	// Intended (instantiated with bindings) KA
  
  KaRuntimeFrame*	runtime_frame;	// KA information related to execution

  IntentionStructure*	intention_structure;	// The IS from whence this came

public:

  ~Goal()
  {
    if (is_toplevel_goal())
      delete goal;

    delete intention;
    delete runtime_frame;
  }

  Goal(GoalAction* ga, Goal* prev=0)
  {
    goal = ga;
    subgoal = 0;
    prev_goal = prev;
    if (prev)
      prev->set_subgoal(this);

    new_goal = 1;

    status = IS_UNTRIED;
    intention = 0;
    runtime_frame = 0;
    intention_structure = 0;
  }

  //

  int generate_soak();

  Binding*	get_goal_binding() const
  {
    return (prev_goal) ? prev_goal->get_intention_binding() : 0;
  }

  Binding*	get_intention_binding() const
  {
    return (intention) ? intention->_binding : 0;
  }

  //
  
  int is_valid() const
  {
    return ((status != IS_SUCCESS) && (status != IS_ABANDONED));
  }

  int is_new() const
  {
    return new_goal;
  }

  int is_toplevel_goal() const
  {
    return (prev_goal == 0);
  }

  int is_leaf_goal() const
  {
    return (subgoal == 0);
  }

  // Check to see if the stack in which this goal is part is blocked.
  int is_stack_blocked();

  // This function should be defined in this goal class
  // because it must use the binding of the subgoaling KA,
  // not the candidate KAs for this goal.
  double eval_priority() const
  {
    Expression* goal_priority = goal->get_priority();

    return goal_priority ? goal_priority->eval(get_goal_binding()).get_real() : 0.0;
  }

  int 	match_relation(Relation* patt_relation, Binding* patt_binding);
  int 	match_goal(GoalAction* goal_action, Binding* goal_action_binding);

  void 	print();


  GoalAction*	get_goal_action() const
  {
    return goal;
  }
  
  const char*	get_name() const
  {
    return (goal) ? goal->get_name() : 0;
  }
  
  Relation*	get_relation() const
  {
    return (goal) ? goal->get_relation() : 0;
  }

  void		set_new()
  {
    new_goal = 1;
  }
  void		clear_new()
  {
    new_goal = 0;
  }

  Goal*		get_subgoal(void) const
  {
    return subgoal;
  }

  void		set_subgoal(Goal* g)
  {
    subgoal = g;
  }
  
  Goal*		get_prev_goal(void) const
  {
    return prev_goal;
  }

  void		set_prev_goal(Goal* g)
  {
    prev_goal = g;
  }

  //

  SoakElement*	get_intention() const
  {
    return intention;
  }

  SoakElement*	set_intention(SoakElement* se);

  IntentionStructure*	get_intention_structure() const
  {
    return (intention_structure) ? intention_structure : 0;
  }

  IntentionStructure*	set_intention_structure(IntentionStructure* is)
  {
    return intention_structure = is;
  }

  StatusType	get_status() const
  {
    return status;
  }
  void		set_status(StatusType st)
  {
    status = st;
  }

  Goal*		get_current_goal(void)
  {
    return intention_structure->get_current_goal();
  }

  void		remove_intention(int failed);

  KaRuntimeFrame* get_runtime_frame(void) const
  {
    return runtime_frame;
  }

  void		set_runtime_frame(KaRuntimeFrame* r)
  {
    if (runtime_frame)
      delete runtime_frame;
    runtime_frame = r;
  }
};


//
//
//

class GoalSet
{

protected:

  Dlist<Goal*>*	_goals;

public:

  ~GoalSet()
  {
    _goals->delete_entities();
    delete _goals;
  }

  GoalSet()
  {
    _goals = new Dlist<Goal*>;
  }

  //
  
  int		all_goals_done();
  
  void		renew_leaf_goals();
  
  void		remove(Goal* gl);
  
  void		drop(GoalAction* ga, Binding* b);
  
  Goal*		add(GoalAction* goal_action, Goal* prev_goal=0);

  Goal* 	add_unique(GoalAction* goal_action, Goal* prev_goal=0, Binding* b=0);

  Goal*		get_goal(long nth) const;

  Dlist<Goal*>*	get_goals()
  {
    return _goals;
  }

  long		size() const
  {
    return _goals->get_count();
  }

public:

  void format(ostream& s, Binding* b);
  void print();
};

typedef Dlist<Goal*>		GoalList;
typedef DlistIterator<Goal*>	GoalListIterator;

#endif
