//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id$
//  $Source$
//  
//  File              : interpreter.cc
//  Author(s)         : Marc Huber <marcush@darwin>
//  
//  Description       : 
//  
//  Original author(s): Marc Huber <marcush@darwin>
//  Organization      : ORINCON Corporation
//  Created On        : Fri Apr 19 15:12:55 1996
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Wed Mar  5 08:13:13 1997
//  Update Count      : 14
//  
//  Copyright (C) 1996 ORINCON Corporation.
//  
//  PURPOSE
//  	|>Description of module's purpose<|
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <iostream.h>

#include "extern.h"
#include "um-prs.h"
#include "is.h"
#include "ka.h"
#include "goal.h"
#include "interpreter.h"
#include "soak.h"

// Forward declaration
class WmTable;


static char kaName[1000];

int Interpreter::interprete(KaTable* ka_library, WmTable* world_model,
			    GoalSet* goal_list, IntentionStructure* is)
{
  Soak*			soak = 0;
  SoakElement*		selected_element;
  long			return_value;

  while (1) {
    is->execute_cycle();

    if (soak)
      delete soak;
    
    soak = new Soak(ka_library, world_model, goal_list);

    // If soak is empty then either no KAs were relevant
    // or there is nothing to do (no new goals).
    if (soak->get_size() == 0) {

      if ((goal_list->all_goals_done())) {
	cout << "\nUM-PRS Exiting.  All system top-level goals achieved!\n";
	delete soak;
	return 0;
      }

      if (is->get_top_level_goal() == 0) {

	// Need to consider the remaining top-level goals
	goal_list->renew_leaf_goals();
	continue;
      }

    }
    else {
    
      if (program.show_soak) {
	cerr << "SOAK generated:\n";
	soak->print(cerr);
      }
      
      // intend one of the KAs selected randomly
      // selected_element = soak->get_random();
      // selected_element = soak->get_priority_first();
      selected_element = soak->get_priority_random();

      if (program.show_soak || program.show_is)
	cerr << "\n\nSelected KA *" << selected_element->_ka->get_name() <<
	  "* from SOAK.\n";

      is->intend(selected_element);

      if(is->get_current_stack() &&
         is->get_current_goal()->get_intention() &&
         is->get_current_goal()->get_intention()->_ka) {
	strcpy(kaName,is->get_current_goal()->get_intention()->_ka->get_name());
      }

    }

    // Go and execute something (in the leaf KA) in the intention structure.
    // If the KA returns a failure then the intention for the active
    // top-level goal has failed and we should clear the IS.
    if (is->activate() == IS_FAILURE)
      is->clear_current_stack();
  }
}

char* getKaName()
{
  return kaName;
}
