//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: is.cc,v 3.1.1.2 1996/02/27 19:39:35 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/is.cc,v $
//  
//  File	      : is.h
//  Author(s)	      : Marcus J. Huber <marcush@engin.umich.edu>
//  
//  Description	      : Intention Structure
//  
//  Original author(s): Marcus J. Huber <marcush@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:26:31 1993
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:50:47 1997
//  Update Count      : 351
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <string.h>

extern "C" {
#include <values.h>
}

#include "extern.h"
#include "um-prs.h"
#include "ka.h"
#include "is.h"
#include "soak.h"
#include "goal.h"

//

const char* const status_desc[] = {
  "IS_UNTRIED",
  "IS_SUCCESS",
  "IS_FAILURE",
  "IS_ACTIVE",
  "IS_BLOCKED",
  "IS_ABANDONED"
};

//
// Check to see if a goal in the stack is blocked
//
int IntentionStack::is_stack_blocked()
{
  Goal*			gl = _top_level_goal;

  while (gl) {
    if (gl->get_status() == IS_BLOCKED)
      return 1;
    gl = gl->get_subgoal();
  }
  return 0;
}


// local function prototypes
int highest_stack_priority(IntentionStackList* _top_level_goals,
			   double priority);


/////////////////////////////////////////////////////////////////////////////
//
// Add a goal to the IS for the system to satisfy.
//
SoakElement* IntentionStructure::intend(SoakElement *element)
{
  SoakElement*			new_se;
  Goal*				old_current;
  Goal*				current_goal;
  Goal*				a_goal;
  int				new_flag = 0;
  IntentionStackListIterator	isli(_top_level_goals);
  IntentionStack*		stack;
  IntentionStack*		current_stack;

  double			element_priority;

  // See if the goal is already in the IS
  current_stack = find_goal_in_stacks(element->_from_goal);

  // Run through all of the stacks and refresh the priorities so
  // that they are current
  refresh_stack_priorities();
  
  // If first top-level goal, then set pointers
  if (_top_level_goals->get_count() == 0) {
    stack = new IntentionStack(element->_from_goal);
    _top_level_goals->append(stack);
    current_goal = element->_from_goal;
    current_goal->set_status(IS_ACTIVE);
    _current_stack = stack;
    new_flag = 1;

    if (program.show_is) {
      print();
    }
    if (program.show_soak) {
      cout << "Added soak element - first goal in intention structure.\n";
    }
  }

  // If this is a top level goal (i.e. no parent goal), or if it is an
  // existing goal (it is a leaf goal in an existing intention stack),
  // we need to see if it has high enough priority to get executed.
  else if ((find_goal_in_stacks(element->_from_goal->get_prev_goal()) == 0) ||
	   (find_current_goal_in_stacks(element->_from_goal) != 0)) {

    // Check priority of soak element and current intentions and run best.  In case
    // of a tie, keep running one of the current intentions.
    current_stack = find_goal_in_stacks(element->_from_goal);
  
    element_priority = element->eval_priority();

    // See if we need to actually add this to the intention structure
    // If the soak element has a higher priority, or if the active
    // stack is currently blocked waiting for a subgoal or a WAIT action,
    // then intend the soak element.
    if (!current_stack &&
	(highest_stack_priority(_top_level_goals, element_priority) ||
	 (_current_stack &&
	  _current_stack->get_current_goal()->get_status() == IS_BLOCKED))) {
      
      // Flag the intention's goal as being ACTIVE
      element->_from_goal->set_status(IS_ACTIVE);
      stack = new IntentionStack(element->_from_goal);
      _top_level_goals->append(stack);
      current_goal = element->_from_goal;
      _current_stack = stack;
      new_flag = 1;

      if (program.show_is) {
	print();
      }
      if (program.show_soak) {
	cout << "Added soak element to intention structure.\n";
      }
    }

    // If the new one doesn't have a high enough priority, then we pick
    // the highest currently intended (preferably unblocked) goal and run it.
    else {
      _current_stack = get_highest_priority_stack();
      // Flag the intention's goal as being ACTIVE (may have been
      // BLOCKED or not set)
      _current_stack->get_current_goal()->set_status(IS_ACTIVE);

      if (program.show_is) {
	print();
      }
      if (program.show_soak) {
	cout << "Did not intend selected soak element - priority too low.\n";
      }

      return element;
    }
  }

  // Otherwise, this is a subgoal
  else {
    _current_stack = current_stack;
    old_current = current_stack->get_current_goal();
    current_goal = current_stack->set_current_goal(element->_from_goal);
  }

  // Need to copy SoakElement element
  new_se = new SoakElement;
  *new_se = *element;
  
  // Add instantiated KA to goal
  current_goal->set_intention_structure(this);
  current_goal->set_intention(new_se);

  // Get KA ready for execution
  if (program.sim_mode) {
    if (element->_ka->get_effect())
      current_goal->set_runtime_frame(element->_ka->get_effect()->get_start_element()->new_runtime_frame());
    else
      current_goal->set_runtime_frame(0);
  }
  else
    current_goal->set_runtime_frame(element->_ka->get_body()->get_start_element()->new_runtime_frame());

/*
  if (current_goal->get_prev_goal())
    current_goal->set_prev_goal(element->_from_goal->get_prev_goal());
    */

  // If a subgoal then duplicate this pointer at the leaf (subgoal) frame.
  if (!new_flag)
    old_current->get_runtime_frame()->intend(new_se);

  if (program.show_is) {
    print();
  }
  if (program.show_soak) {
    cout << "Just intended new SoakElement.\n";
  }
  
  return new_se;
}


/////////////////////////////////////////////////////////////////////////////
//
// Go through all the stacks in the IS and update every intention's priority
//
void IntentionStructure::refresh_stack_priorities()
{
  IntentionStackListIterator	isli(_top_level_goals);
  IntentionStack*		stack;
  Goal*				stack_goal;

  for(isli.reset(), stack = isli.get_next(); stack; stack = isli.get_next()) {
    stack_goal = stack->get_top_level_goal();
    
    while (stack_goal) {
      SoakElement* intent;
      if (intent = stack_goal->get_intention())
	intent->_ka->confirm_context(stack_goal->get_intention_binding());
      stack_goal = stack_goal->get_subgoal();
    }
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// Search for and return the highest priority (preferably unblocked)
// intention stack
//
IntentionStack* IntentionStructure::get_highest_priority_stack() const
{
  IntentionStackListIterator	isli1(_top_level_goals);
  IntentionStack*		stack;
  double			highest_priority;
  double			highest_unblocked_priority;
  IntentionStack*		highest_stack = 0;
  IntentionStack*		highest_unblocked_stack = 0;
  

  // Compare everything to the currently executing stack
  highest_stack = _current_stack;
  if (highest_stack) {
    highest_priority = highest_stack->get_current_goal()->get_intention()->eval_priority();
    if (highest_stack->get_current_goal()->get_status() != IS_BLOCKED) {
      highest_unblocked_priority = highest_priority;
      highest_unblocked_stack = highest_stack;
    }
  }
  else {
    highest_priority = highest_unblocked_priority = -MAXDOUBLE;
  }

  isli1.reset();
  stack = isli1.get_next();

  while (stack) {

    double stack_priority;
    if ((stack_priority = stack->get_current_goal()->get_intention()->eval_priority()) >
	highest_priority) {
      highest_priority = stack_priority;
      highest_stack = stack;
    }

    if ((stack->get_current_goal()->get_status() != IS_BLOCKED) &&
	(stack_priority > highest_unblocked_priority)) {
      highest_unblocked_priority = stack_priority;
      highest_unblocked_stack = stack;
    }

    stack = isli1.get_next();
  }
  
  return (highest_unblocked_stack) ? highest_unblocked_stack : highest_stack;
}


/////////////////////////////////////////////////////////////////////////////
//
// Search for and remove the indicated intention stack
//
void IntentionStructure::remove_stack(IntentionStack* is)
{
  IntentionStackListIterator	isli1(_top_level_goals);
  IntentionStack*		stack;
  
  isli1.reset();
  stack = isli1.get_next();

  while (stack && stack != is) {
    stack = isli1.get_next();
  }

  if (stack) {
    isli1.remove_this();
    if (stack == _current_stack)
      _current_stack = 0;
    delete stack;
  }
  else {
    cerr << "IS::remove_stack - couldn't find stack!\n";
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// Remove the current stack from the Intention Structure
//
void IntentionStructure::clear_current_stack(void)
{
  if (_current_stack) {
    _current_stack->get_top_level_goal()->remove_intention(0);
    remove_stack(_current_stack);
    _current_stack = get_highest_priority_stack();
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// Try to run or subgoal something in a KA body
//
StatusType IntentionStructure::activate(void)
{
  Goal*				top_level_goal;
  Goal*				new_goal;
  KaRuntimeFrame*		top_level_frame;
  KaBodyElement::KaReturnType	return_val;
  Binding*			b;
  KaBodyAtomicElement*		failure_section;

  // Check to see if we have anything intended on the IS
  if (_current_stack == 0) {
    cerr << "Nothing currently active.\n";
    return IS_SUCCESS;
  }
  
  top_level_goal = _current_stack->get_top_level_goal();
  top_level_frame = top_level_goal->get_runtime_frame();

  // If we're trying to run a top level goal with a non-existant
  // EFFECTS section, we need to mark the goal as having succeeded
  // and remove it from the IS and system goal list
  if (program.sim_mode && (top_level_frame == 0)) {
    top_level_goal->set_runtime_frame(0);
    program.goal_list->remove(top_level_goal);
    remove_stack(_current_stack);
    _current_stack = get_highest_priority_stack();
    return IS_SUCCESS;
  }

  // At this point we'll be executing an action from a KA so we have
  // to check the top goal to see if it's still valid.
  if (!top_level_goal->get_intention()->_ka->confirm_context(top_level_goal->get_intention_binding()) ||
      (top_level_goal->get_status() == IS_ABANDONED)) {

    // OOPS!  top level goal invalid!
    top_level_goal->remove_intention(1);
    if (top_level_goal->get_status() == IS_ABANDONED) {
      program.goal_list->remove(top_level_goal);
    }
    else {
      top_level_goal->set_status(IS_FAILURE);
      top_level_goal->set_subgoal(0);
    }

    _current_stack->set_current_goal(0);
    program.goal_list->renew_leaf_goals();
    return IS_FAILURE;
  }
  
  return_val = top_level_frame->execute(top_level_goal->get_intention_binding(), top_level_goal);
  if (return_val == KaBodyElement::KA_ELEMENT_FAILED) {
    
    if (program.show_is || program.show_gl)
      cout << "Top-level goal " << top_level_goal->get_name() << " failed!\n";
    
    // Execute the FAILURE section of the KA here
    if ((failure_section = top_level_goal->get_intention()->_ka->get_failure()) != 0) {
      top_level_frame = failure_section->new_runtime_frame();
      top_level_frame->execute(top_level_goal->get_intention_binding(), top_level_goal);
    }

    top_level_goal->remove_intention(0);
    top_level_goal->set_status(IS_FAILURE);
    top_level_goal->set_subgoal(0);
    program.goal_list->renew_leaf_goals();
    return IS_FAILURE;
  }
  else if (return_val == KaBodyElement::KA_ELEMENT_INCOMP) {
    return IS_ACTIVE;
  }
  else /* if (return_val == KaBodyElement::KA_ELEMENT_COMPLETE) */ {
    
    // Done with the KA!
    if (top_level_frame->get_this_element()->is_final_element()) {
      top_level_goal->set_runtime_frame(0);
      program.goal_list->remove(top_level_goal);
      remove_stack(_current_stack);
      _current_stack = get_highest_priority_stack();
      program.goal_list->renew_leaf_goals();
      return IS_SUCCESS;
    }
    else {
      top_level_goal->set_runtime_frame(top_level_frame->get_this_element()->get_next_element()->new_runtime_frame());
    }
    return IS_ACTIVE;
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// Execute the global Cyclic procedure
//
StatusType IntentionStructure::execute_cycle(void)
{
  KaRuntimeFrame*		top_level_frame;
  KaBodyElement::KaReturnType	return_val;

  if (program.cycle != 0) {
    Binding* b = new Binding(program.cycle->get_symtab());
    
    top_level_frame = program.cycle->get_body()->get_start_element()->new_runtime_frame();

    while (top_level_frame) {

      return_val = top_level_frame->execute(b, 0);
      if (return_val == KaBodyElement::KA_ELEMENT_FAILED) {
	cout << "Cycle failed!\n";
	delete top_level_frame;
	delete b;
	return IS_FAILURE;
      }
      else if (return_val == KaBodyElement::KA_ELEMENT_COMPLETE) {
	if (top_level_frame->get_this_element()->is_final_element()) {
	  delete top_level_frame;
	  delete b;
	  return IS_SUCCESS;
	}
	else {
	  KaRuntimeFrame* old_frame = top_level_frame;
	  top_level_frame = top_level_frame->get_this_element()->get_next_element()->new_runtime_frame();
	  delete old_frame;
	}
      }
      else if (return_val == KaBodyElement::KA_ELEMENT_INCOMP) {
	continue;
      }
    }
    delete top_level_frame;
    delete b;
  }
}


/////////////////////////////////////////////////////////////////////////////
//
// Return the intention structure stack in which the goal is found
//
IntentionStack* IntentionStructure::find_goal_in_stacks(Goal* goal)
{
  IntentionStackListIterator	isli(_top_level_goals);
  IntentionStack*		stack;
  Goal*				stack_goal;

  for(isli.reset(), stack = isli.get_next(); stack; stack = isli.get_next()) {
    stack_goal = stack->get_top_level_goal();
    
    while (stack_goal) {
      if (goal == stack_goal)
	return stack;
      stack_goal = stack_goal->get_subgoal();
    }
  }
  
  return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
// Return the intention structure stack in which the top level goal is found
//
IntentionStack* IntentionStructure::find_toplevel_goal_in_stacks(Goal* goal)
{
  IntentionStackListIterator	isli(_top_level_goals);
  IntentionStack*		stack;
  Goal*				stack_goal;

  for(isli.reset(), stack = isli.get_next(); stack; stack = isli.get_next()) {
    stack_goal = stack->get_top_level_goal();
    
    if (goal == stack_goal)
	return stack;
  }
  
  return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
// Return the intention structure stack in which the goal is found as a
// current goal in the stack.
//
IntentionStack* IntentionStructure::find_current_goal_in_stacks(Goal* goal)
{
  IntentionStackListIterator	isli(_top_level_goals);
  IntentionStack*		stack;
  Goal*				stack_goal;

  for(isli.reset(), stack = isli.get_next(); stack; stack = isli.get_next()) {
    stack_goal = stack->get_current_goal();
    
    if (goal == stack_goal)
	return stack;
  }
  
  return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
// Display the current state of the IntentionStructure
//
void IntentionStructure::print()
{
  IntentionStackListIterator	isli(_top_level_goals);
  IntentionStack*		stack;
  
  Goal*		active_goal;
  long		intent_num = 0;

  cout << "\nCurrent IntentionStructure state:\n";
  if (_current_stack)
    cout << "Current stack => " <<
      _current_stack->get_top_level_goal()->get_name() <<
      "\n";

  for(intent_num = 0, isli.reset(), stack = isli.get_next();
      stack;
      stack = isli.get_next(), intent_num++) {
    
    cout << form("\nIntentionStack#%d: (first entry is top_level_goal)\n",
		 intent_num);
    
    // Chain down IS until we run out of intentions
    active_goal = stack->get_top_level_goal();
    
    while (active_goal) {
      active_goal->print();
      active_goal = active_goal->get_subgoal();
    }
  }
}


//
//

/////////////////////////////////////////////////////////////////////////////
//
// Return 1 if the given priority is higher than all those in the Intention
// stacks (i.e. higher than any stack's leaf-level goal's priority)
//
int highest_stack_priority(IntentionStackList* _top_level_goals,
			   double priority)
{
  IntentionStackListIterator	isli(_top_level_goals);
  IntentionStack*		stack;
  SoakElement*			intention;
  Goal*				stack_goal;
  double			intention_priority;
  
  for(isli.reset(), stack = isli.get_next(); stack; stack = isli.get_next()) {

    stack_goal = stack->get_current_goal();

    // Update the variable bindings with the latest values
//    stack_goal->get_intention()->_ka->confirm_context(stack_goal->get_intention_binding());

    intention = stack_goal->get_intention();
    intention_priority = intention->eval_priority();

    if (program.show_is) {
      cout << form("  Intention priority %5.2f, intended element priority %5.2f\n",
		   intention_priority, priority);
    }
    
    if (intention_priority >= priority)
      return 0;
  }
  
  return 1;
}

//
//

const char *indent(class ostream& stream, long depth)
{
  int i;
  const char a_space[2] = " ";

//  stream << depth;
  for (i = 0; i < depth && i < 40; i++)
    stream << "	 ";
  return a_space;
}


