//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: is.h,v 3.1.1.2 1996/02/27 19:39:35 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/is.h,v $
//  
//  File	      : is.h
//  Author(s)	      : Marcus J. Huber <marcush@engin.umich.edu>
//  
//  Description	      : Intention Structure
//  
//  Original author(s): Marcus J. Huber <marcush@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:26:31 1993
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:51:45 1997
//  Update Count      : 179
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __IS_H__
#define __IS_H__

#include "dlist.h"

// forward declarations

class SoakElement;

class Goal;

class IntentionStructure;

const char *indent(class ostream& stream=cout, long depth=0);

//

enum StatusType
{
  IS_UNTRIED,
  IS_SUCCESS,
  IS_FAILURE,
  IS_ACTIVE,
  IS_BLOCKED,
  IS_ABANDONED
};

extern const char* const status_desc[];

  
/////////////////////////////////////////////////////////////////////////////
//
class IntentionStack
{
  Goal*		_top_level_goal;	// The system goal being pursued.
  Goal*		_current_goal;		// The goal that is actually being executed.
  
public:

  IntentionStack()
  {
    _top_level_goal = 0;
    _current_goal   = 0;
  };
  
  IntentionStack(Goal* g)
  {
    _top_level_goal = g;
    _current_goal   = g;
  };
  
  // access functions
  Goal*		get_top_level_goal() const	{ return _top_level_goal; }
  Goal*		set_top_level_goal(Goal* g) 	{ return _top_level_goal = g; }
  
  Goal*		get_current_goal() const	{ return _current_goal; }
  Goal*		set_current_goal(Goal* g) 	{ return _current_goal = g; }

  int		is_stack_blocked();
};


typedef Dlist<IntentionStack*>		IntentionStackList;
typedef DlistIterator<IntentionStack*>	IntentionStackListIterator;



/////////////////////////////////////////////////////////////////////////////
//
class IntentionStructure
{
  IntentionStackList*	_top_level_goals;	// The system goals being pursued.
  IntentionStack*	_current_stack;		// the active stack
  
public:

  IntentionStructure()
  {
    _top_level_goals = new IntentionStackList;
    _current_stack   = 0;
  }

  ~IntentionStructure()
  {
    delete _top_level_goals;
  }
  
  SoakElement*		intend(SoakElement *k);
  void			clear_current_stack(void);
  StatusType		activate(void);
  StatusType		execute_cycle(void);
  void			print();

  IntentionStack*	find_goal_in_stacks(Goal* goal);
  IntentionStack*	find_toplevel_goal_in_stacks(Goal* goal);
  IntentionStack*	find_current_goal_in_stacks(Goal* goal);
  
  Goal*			get_current_goal()
  {
    return _current_stack->get_current_goal();
  }
  
  Goal*			set_current_goal(Goal* goal)
  {
    return _current_stack->set_current_goal(goal);
  }
  
  Goal*			get_top_level_goal()
  {
    return (_current_stack) ? _current_stack->get_top_level_goal() : 0;
  }
  
  // access functions
  IntentionStack*	get_current_stack() const	{ return _current_stack; }
  IntentionStack*	set_current_stack(IntentionStack* is)
  {
    return _current_stack = is;
  }

  void			refresh_stack_priorities();
  IntentionStack*	get_highest_priority_stack() const;
  
  void			remove_stack(IntentionStack* is);
  
  IntentionStackList*	get_top_level_goals() const	{ return _top_level_goals; }
};

#endif
