//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//
//  $Id: ka.cc,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/ka.cc,v $
//
//  File	      : ka.cc
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//
//  Description	      : Knowledge Area
//
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:24:34 1993
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:52:46 1997
//  Update Count      : 303
//
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
//
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>

#include "extern.h"
#include "ka.h"
#include "relation.h"
#include "goal.h"
#include "run.h"
#include "is.h"

//
// KaContext
//

Ka::~Ka()				
{
  delete _symtab;
  delete _name;
  delete _documentation;
  delete _purpose;
  delete _context;
  delete _priority;
  delete _effect;
  delete _failure;
  delete _body;
}

int Ka::check_context(BindingList& bl) const
{
  return _context ? _context->check(bl) : 1;
}

int Ka::confirm_context(Binding* b) const
{
  return _context ? _context->confirm(b) : 1;
}


//

void KaTable::disable(const char* name)
{
  KaTableIterator next_ka(this);
  Ka* ka;

  while(ka = next_ka())
    if (!strcmp(ka->get_name(), name))
    ka->disable();
}

//
// KaContext
//

int KaContext::check(BindingList& bl) const
// establish contexts (generate the binding list)
{
  ConditionListIterator next_condition(_conditions);
  Condition* condition;

  while(condition = next_condition()) {

#ifdef DEBUG
    cerr << form("[KaContext::check] (condition) \n");
    cerr << "Type : " << condition->type() << ",\t"
      << "Name : " << condition->get_name() << "\n";
#endif

    if (!condition->check(bl))
      break;
  }

  return bl.get_count();
}

int KaContext::confirm(Binding* b) const
// confirm the validity of the current context with the current binding
{
  ConditionListIterator next_condition(_conditions);
  Condition* condition;

  while(condition = next_condition())
    if (!condition->confirm(b)) return 0;

  return 1;
}

//
// KaBodyElement
//


//
// KaBodySimpleElement
//

KaRuntimeFrame*	KaBodySimpleElement::new_runtime_frame()
{

  if (action->is_executable_action())
    return new KaSimpleRuntimeFrame(this);
  else
    return new KaGoalRuntimeFrame(this);
}


//
// KaBodyBranchElement
//

KaBodyBranchElement::~KaBodyBranchElement()
{
  for (long i=0; i < num_branches; i++) {
    KaBodyElement* an_element = branch_start[i];
    KaBodyElement* next;

    while (an_element) {
      next = an_element->get_next_element();
      delete an_element;
      an_element = next;
    }
  }
  delete[] branch_start;
}

void KaBodyBranchElement::add_branch(KaBodyElement* e)
{
  KaBodyElement** nb = new KaBodyElement*[num_branches+1];

  for (long i = 0; i < num_branches; i++)
    nb[i] = branch_start[i];

  delete[] branch_start;
  branch_start = nb;
  branch_start[num_branches] = e;
  num_branches++;
}


KaRuntimeFrame*	KaBodyBranchElement::new_runtime_frame()
{
  KaBranchRuntimeFrame* tmp;
  tmp = new KaBranchRuntimeFrame(this);
  return tmp;
}

//
// KaBodyWhenElement
//

KaBodyWhenElement::~KaBodyWhenElement()
{
  delete test;
  
  KaBodyElement* an_element = start_element;
  KaBodyElement* next;

  while (an_element) {
    next = an_element->get_next_element();
    delete an_element;
    an_element = next;
  }
}

KaRuntimeFrame*	KaBodyWhenElement::new_runtime_frame()
{
  KaWhenRuntimeFrame* tmp;
  tmp = new KaWhenRuntimeFrame(this);
  return tmp;
}


//
// KaBodyWhileElement
//

KaBodyWhileElement::~KaBodyWhileElement()
{
  delete test;
  
  KaBodyElement* an_element = start_element;
  KaBodyElement* next;

  while (an_element) {
    next = an_element->get_next_element();
    delete an_element;
    an_element = next;
  }
}

KaRuntimeFrame*	KaBodyWhileElement::new_runtime_frame()
{
  KaWhileRuntimeFrame* tmp;
  tmp = new KaWhileRuntimeFrame(this);
  return tmp;
}


//
// KaBodyDoElement
//

KaBodyDoElement::~KaBodyDoElement()
{
  delete test;
  
  KaBodyElement* an_element = start_element;
  KaBodyElement* next;

  while (an_element) {
    next = an_element->get_next_element();
    delete an_element;
    an_element = next;
  }
}

KaRuntimeFrame*	KaBodyDoElement::new_runtime_frame()
{
  KaDoRuntimeFrame* tmp;
  tmp = new KaDoRuntimeFrame(this);
  return tmp;
}


//
// KaBodyAtomicElement
//

KaBodyAtomicElement::~KaBodyAtomicElement()
{
  KaBodyElement* an_element = start_element;
  KaBodyElement* next;

  while (an_element) {
    next = an_element->get_next_element();
    delete an_element;
    an_element = next;
  }
}

KaRuntimeFrame*	KaBodyAtomicElement::new_runtime_frame()
{
  KaAtomicRuntimeFrame* tmp;
  tmp = new KaAtomicRuntimeFrame(this);
  return tmp;

  // Try this out again once things are working
//  return new KaAtomicRuntimeFrame(this);
}


//
// KaBody
//

KaBody::~KaBody()
{
  KaBodyElement* an_element = start_element;
  KaBodyElement* next;

  while (an_element) {
    next = an_element->get_next_element();
    delete an_element;
    an_element = next;
  }
}


//
// KaTableBucketIterator
//

KaTableBucketIterator::KaTableBucketIterator(KaTable* kt, Goal* goal)
{
  _table = kt;
  _goal = goal;
  _bucket = new SymbolListIterator(kt->get_bucket(goal->get_name()));
}

  
Ka* KaTableBucketIterator::operator()(Binding*& ka_binding)
{	
  KaEnt* ke = (KaEnt*) _bucket->operator()();

  if (!ke) return 0;

  Ka* ka = ke->get_ka();
  ka_binding = new Binding(ka->get_symtab());

  if (!_goal->get_goal_action()->is_eligible(ka->get_name(), ka_binding)) {
    delete ka_binding;
    return operator()(ka_binding);
  }
					
  Relation* ka_relation =  ka->get_purpose()->get_relation();

  if (_goal->match_relation(ka_relation, ka_binding))
    return ka;
  else {
    delete ka_binding;
    return operator()(ka_binding);
  }
}
