//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: predicate.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/predicate.h,v $
//  
//  File              : predicate.h
//  Author(s)         : Jaeho Lee <jaeho@az.engin.umich.edu>
//  
//  Description       : predicate
//  
//  Original author(s): Jaeho Lee <jaeho@az.engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On        : Tue Jul 12 22:21:42 1994
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:06:49 1996
//  Update Count      : 127
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __PREDICATE__
#define __PREDICATE__

#include <string.h>
#include "exp.h"
#include "exp-val.h"
#include "relation.h"
#include "wm.h"
#include "goal.h"

//
// Predicate
//

class Predicate : public Expression
{
  char* 	_name;
  Relation*	_relation;

  void init(char* name, Relation* relation=0)
  {
    ::strcpy(_name=new char[::strlen(name)+1], name);
    _relation = relation;
  }

public:
  virtual ~Predicate()
  {
    delete _name;
    delete _relation;
  }
  
  Predicate(char* name)
  {
    init(name);
  }

  Predicate(char* name, Relation* relation)
  {
    init(name, relation);
  }

  // access functions
  const char* get_name() const
  {
    return _name;
  }
  
  const ExpType get_type() const
  {
    return EXP_PREDICATE;
  }
  
  Relation* get_relation() const
  {
    return _relation;
  }
  
  const int is_variable() const
  {
    return 0;
  }

  virtual Value eval(Binding*const binding=0) = 0;

public:
  
  virtual void print(ostream& s=cout, Binding* b=0)
  {
    s << "Name: " << get_name()
      << ",\tValue = " << (eval(b) != 0 ? "True" : "False") << "\n";
  }

  virtual void format(ostream& s=cout, Binding* b=0)
  {
    s << '(';
    s << get_name() << " ";
    _relation->format(s, b);
    s << ')';
  }
};

//

class PredicateFact : public Predicate
{

public:
  
  PredicateFact(char* name, Relation* relation)
  : Predicate(name, relation)
  {}

  Value eval(Binding*const binding=0)
  {
    if (program.world_model->match(get_relation(), binding))
      return Value::True;
    else
      return Value::False;
  }   
};

//

class PredicateRetrieve : public Predicate
{

public:
  
  PredicateRetrieve(char* name, Relation* relation)
  : Predicate(name, relation)
  {}

  Value eval(Binding*const binding=0)
  {
    // unbound variables before matching
   binding->unbind_variables(get_relation()->get_args());
    
    // now try match to get new values
   if (program.world_model->match(get_relation(), binding))
     return Value::True;
   else
     return Value::False;
  }   
};

//

class PredicateAchieve : public Predicate
{

public:
  
  PredicateAchieve(char* name, Relation* relation)
  : Predicate(name, relation)
  {}

  Value eval(Binding*const binding=0)
  {
    register Goal*	goal;
    GoalListIterator next_goal(program.goal_list->get_goals());

    while(goal = next_goal())
      if (goal->match_relation(get_relation(), binding))
	return Value::True;

    return Value::False;
  }   
};

#endif
