//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: primitive.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/primitive.cc,v $
//  
//  File	      : primitive.cc
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : primitive functions
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Mon Jan 31 10:12:39 1994
//  Last Modified By  : Marcus J. Huber <marcush@ultra>
//  Last Modified On  : Fri Feb 21 12:54:40 1997
//  Update Count      : 225
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>
#include <unistd.h>
#include <sys/time.h>

#include "extern.h"
#include "binding.h"
#include "symtab.h"
#include "wm.h"
#include "exp.h"
#include "primitive.h"

extern "C"
{
  extern system(char*);
}

//
// The KA function list is at the end of this file
//


PRIMITIVE_FUNCTION(primitive_noop)
{
  return Value::True;
}


PRIMITIVE_FUNCTION(primitive_plus)
{
  if (arity <= 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();
  
  Value result = exp->eval(binding);

  while (exp = next_exp())
    // don't use += (it is not defined for Value)
    result = result + exp->eval(binding);
  
  return result;
}


PRIMITIVE_FUNCTION(primitive_minus)
{
  if (arity <= 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value result = exp->eval(binding);

  if (arity == 1) 
    return -result;

  while (exp = next_exp())
    // don't use -= (it is not defined for Value)
    result = result - exp->eval(binding);
  
  return result;
}

PRIMITIVE_FUNCTION(primitive_mul)
{
  if (arity <= 0) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();
  
  Value result = exp->eval(binding);

  while (exp = next_exp())
    // don't use *= (it is not defined for Value)
    result = result * exp->eval(binding);
  
  return result;
}

PRIMITIVE_FUNCTION(primitive_div)
{
  if (arity <= 0) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value result = exp->eval(binding);

  if (arity == 1) 
    return Value(1) / result;

  while (exp = next_exp())
    // don't use /= (it is not defined for Value)
    result = result / exp->eval(binding);
  
  return result;
}

PRIMITIVE_FUNCTION(primitive_mod)
{
  if (arity <= 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value result = exp->eval(binding);

  while (exp = next_exp())
    // don't use %= (it is not defined for Value)
    result = result % exp->eval(binding);
  
  return result;
}

PRIMITIVE_FUNCTION(primitive_abs)
{
  if (arity < 1) return Value::False;

  Value value(args->first()->eval(binding));

  return (value < 0) ? -value : value;
}

//

PRIMITIVE_FUNCTION(test_eq)
{
  if (arity <= 1) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value oldval, newval;

  oldval = exp->eval(binding);
  while (exp = next_exp()) {
    newval = exp->eval(binding);
    if (oldval != newval) return Value::False;
    oldval = newval;
  }

  return Value::True;
}

//

PRIMITIVE_FUNCTION(primitive_eq)
{
  return test_eq(arity, args, binding);
}

PRIMITIVE_FUNCTION(primitive_neq)
{
  return ! test_eq(arity, args, binding);
}


PRIMITIVE_FUNCTION(primitive_lt)
{
  if (arity <= 1) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value oldval, newval;

  oldval = exp->eval(binding);
  while (exp = next_exp()) {
    newval = exp->eval(binding);
    if (oldval >= newval) return Value::False;
    oldval = newval;
  }
  return Value::True;
}

PRIMITIVE_FUNCTION(primitive_le)
{
  if (arity <= 1) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value oldval, newval;

  oldval = exp->eval(binding);
  while (exp = next_exp()) {
    newval = exp->eval(binding);
    if (oldval > newval) return Value::False;
    oldval = newval;
  }
  return Value::True;
}

PRIMITIVE_FUNCTION(primitive_gt)
{
  if (arity <= 1) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value oldval, newval;

  oldval = exp->eval(binding);
  while (exp = next_exp()) {
    newval = exp->eval(binding);
    if (oldval <= newval) return Value::False;
    oldval = newval;
  }
  return Value::True;
}

PRIMITIVE_FUNCTION(primitive_ge)
{
  if (arity <= 1) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp = next_exp();

  Value oldval, newval;

  oldval = exp->eval(binding);
  while (exp = next_exp()) {
    newval = exp->eval(binding);
    if (oldval < newval) return Value::False;
    oldval = newval;
  }
  return Value::True;
}

//

PRIMITIVE_FUNCTION(primitive_and)
{
  if (arity <= 0) return Value::True;

  ExpListIterator next_exp(args);

  for (Expression* exp = next_exp(); exp; exp = next_exp())
    if (!(exp->eval(binding).is_true())) return Value::False;
  
  return Value::True;
}

PRIMITIVE_FUNCTION(primitive_or)
{
  if (arity <= 0) return Value::True;

  ExpListIterator next_exp(args);

  for (Expression* exp = next_exp(); exp; exp = next_exp())
    if (exp->eval(binding).is_true()) return Value::True;
  
  return Value::False;
}

PRIMITIVE_FUNCTION(primitive_not)
{
  if (arity != 1) return Value::False;

  ExpListIterator next_exp(args);

  return (!(next_exp()->eval(binding).is_true()));
}


//
//
//

PRIMITIVE_FUNCTION(primitive_is_defined)
{
  if (arity < 1) return Value::False;
							
  return (args->first()->eval(binding).is_defined());
}

PRIMITIVE_FUNCTION(primitive_is_void)
{
  if (arity < 1) return Value::False;
							
  return (args->first()->eval(binding).get_type() == Value::VAL_VOID);
}


PRIMITIVE_FUNCTION(primitive_is_integer)
{
  if (arity < 1) return Value::False;
							
  return (args->first()->eval(binding).get_type() == Value::VAL_INT);
}

PRIMITIVE_FUNCTION(primitive_is_real)
{
  if (arity < 1) return Value::False;
							
  return (args->first()->eval(binding).get_type() == Value::VAL_REAL);
}

PRIMITIVE_FUNCTION(primitive_is_string)
{
  if (arity < 1) return Value::False;

  return (args->first()->eval(binding).get_type() == Value::VAL_STRING);
}

//
//
//

PRIMITIVE_FUNCTION(primitive_print)
{
  if (arity <= 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp;

  for(exp = next_exp(); exp; exp = next_exp())
    exp->eval(binding).print(cout, binding);

  return Value::True;
}


PRIMITIVE_FUNCTION(primitive_create_relation)
{
  if (arity <= 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp;

  const char* rel_name = next_exp()->eval(binding).get_string();
  ExpList* rel_args = new ExpList;

  for(exp = next_exp(); exp; exp = next_exp())
    rel_args->append(new Value(exp->eval(binding)));

  program.world_model->add(new WmRelation(rel_name, rel_args));

  return Value::True;
}


PRIMITIVE_FUNCTION(primitive_sleep)
{
  if (arity < 1) return Value::False;

  ExpListIterator next_exp(args);

  sleep((int) next_exp()->eval(binding));

  return Value::True;
}


PRIMITIVE_FUNCTION(primitive_shell)
{
  if (arity < 1) return Value::False;

  ExpListIterator next_exp(args);

  if (system(next_exp()->eval(binding)) < 0)
    return Value::False;
  else
    return Value::True;
}


PRIMITIVE_FUNCTION(primitive_get_time)
{
  if (arity < 1) return Value::False;
							
  Expression* exp = args->first();
							
  if (!exp->is_variable())
    return Value::False;

  long t = time(0);
  Value tt(t);
  binding->set_value(exp, tt);

  return Value::True;					
}


PRIMITIVE_FUNCTION(primitive_read_integer)
{
  if (arity < 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp;
  long response;
  
  for(exp = next_exp(); exp; exp = next_exp()) {
    if (!exp->is_variable())
      return Value::False;
      
    cin >> response;
    Value value(response);
    binding->set_value(exp, value);
  }
  
  return Value::True;
}


PRIMITIVE_FUNCTION(primitive_read_float)
{
  if (arity < 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp;
  double response;
  
  for(exp = next_exp(); exp; exp = next_exp()) {
    if (!exp->is_variable())
      return Value::False;
      
    cin >> response;
    Value value(response);
    binding->set_value(exp, value);
  }
  
  return Value::True;
}


PRIMITIVE_FUNCTION(primitive_read_string)
{
  if (arity < 0) return Value::False;

  ExpListIterator next_exp(args);
  Expression* exp;
  char response[READ_STRING_INPUT_SIZE];
  
  for(exp = next_exp(); exp; exp = next_exp()) {
    cin >> response;

    Value value(response);
    binding->set_value(exp, value);
  }
  
  return Value::True;
}


#ifdef UM_PRS
#include "ka.h"
#include "soak.h"

PRIMITIVE_FUNCTION(primitive_remove_ka)
{
  if (arity <= 0) return Value::True;

  ExpListIterator next_exp(args);
  Expression* exp;
  
  for(exp = next_exp(); exp; exp = next_exp())
    program.ka_library->disable(exp->eval(binding));
  
  return Value::True;
}

PRIMITIVE_FUNCTION(primitive_generate_soak)
{
  if (arity < 1) return 0;
							
  Expression* exp = args->first();
							
  if (!exp->is_variable())
    return 0;

  Soak*	s = new Soak(program.ka_library, program.world_model, program.goal_list, 1);
  
  binding->set_value(exp, Value((void*)s));

  return s->get_size();
}

PRIMITIVE_FUNCTION(primitive_delete_soak)
{
  if (arity < 1) return Value::False;
							
  Expression* exp = args->first();
							
  if (!exp->is_variable())
    return Value::False;

  delete (Soak*) exp->eval(binding).get_void();

  return Value::True;
}

PRIMITIVE_FUNCTION(primitive_print_soak)
{
  if (arity < 1) return Value::False;
							
  Expression* exp = args->first();
							
  if (!exp->is_variable())
    return Value::False;

  Soak*	soak = (Soak*) exp->eval(binding).get_void();

  soak->print(cerr);
  cerr << "\n";

  return Value::True;
}

#endif

////
////
////
////

Function lookup_function_table(const char* name)
{
  FunctionNamePair* fnp;

  for (fnp = user_function_list; fnp->fp; fnp++)
    if (strcmp(fnp->name, name) == 0) return fnp->fp;

  for (fnp = system_function_list; fnp->fp; fnp++)
    if (strcmp(fnp->name, name) == 0) return fnp->fp;
  
  return 0;
}

//
//
//

FunctionNamePair system_function_list[] = {
  {primitive_noop		,	"noop"			},

  {primitive_plus		,	"+"			},
  {primitive_minus		,	"-"			},
  {primitive_mul		,	"*"			},
  {primitive_div		,	"/"			},
  {primitive_mod		,	"%"			},
  {primitive_abs		,	"abs"			},  
  
  {primitive_eq			,	"=="			},
  {primitive_neq		,	"!="			},
  {primitive_lt			,	"<"			},
  {primitive_le			,	"<="			},
  {primitive_gt			,	">"			},
  {primitive_ge			,	">="			},
			
  {primitive_and		,	"and"			},
  {primitive_or			,	"or"			},
  {primitive_not		,	"not"			},
			  
  {primitive_is_defined		,	"is_defined"		},
  {primitive_is_void		,	"is_void"     		},
  {primitive_is_integer		,	"is_integer"  		},
  {primitive_is_real		,	"is_real"     		},
  {primitive_is_string		,	"is_string"   		},

  {primitive_and		,	"&&"			},
  {primitive_or			,	"||"			},
  {primitive_not		,	"!"			},
  {primitive_print		,	"print"			},
  {primitive_create_relation	,	"create_relation"	},
  {primitive_sleep		,	"sleep"			},
  {primitive_shell		,	"shell"			},
  {primitive_get_time		,	"get_time"		},

  {primitive_read_integer	,	"read_integer"		},
  {primitive_read_float		,	"read_float"		},
  {primitive_read_string	,	"read_string"		},

#ifdef UM_PRS
  {primitive_remove_ka	,		"remove_ka"		},
  {primitive_generate_soak	,	"generate_soak"		},
  {primitive_delete_soak	,	"delete_soak"		},
  {primitive_print_soak		,	"print_soak"		},
#endif

  {0				,	0			}
};
