//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: random.h,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/random.h,v $
//  
//  File	      : random.h
//  Author(s)	      : Jaeho Lee <jaeho@dworkin.eecs.umich.edu>
//  
//  Description	      : 
//  
//  Original author(s): Jaeho Lee <jaeho@dworkin.eecs.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Fri Aug 20 15:05:45 1993
//  Last Modified By  : Jaeho Lee <jaeho@splat.engin.umich.edu>
//  Last Modified On  : Mon Feb 26 22:50:22 1996
//  Update Count      : 210
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////


#ifndef __RANDOM_H__
#define __RANDOM_H__

#define RAND48

//

#ifdef	RAND48

#include <stdlib.h>
#include <values.h>
#include <math.h>
#include <time.h>

extern "C"
{
  void 		srand48(long seedval);
  long 		lrand48();
  double	drand48();
}

#define	RANDOM_INTEGER	lrand48()

#else

extern "C"
{
#include <stdlib.h>
#include <values.h>
#include <math.h>
#include <time.h>

  extern long 	random(...);
  extern char* 	initstate(...);
  extern char* 	setstate(...);
}

static long state1[32] = {
  3,
  0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342,
  0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd,
  0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86,
  0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7,
  0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc,
  0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
  0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b,
  0xf5ad9d0e, 0x8999220b, 0x27fb47b9
};

#define	RANDOM_INTEGER	random()

#endif


//

static void initialize_generator() 
{
  static initialized = 0;
  time_t seed;
  
  if (!initialized) {
    time(&seed);

#ifdef	RAND48
    srand48((long) seed);
#else
    initstate((unsigned) seed, (char*) state1, 128);
    setstate((char*) state1);
#endif
    initialized = 1;
  }
}

//
//
//

class Random_Uniform_Integer
// generate integer random numbers between min and max range
{
  long _minimum;
  long _range;

public:

  void set_parameter(long maximum=MAXLONG, long minimum=0)
  {
    _minimum = minimum;
    _range = maximum - minimum;
  }

  // Will generate number from min (inclusive) to max (exclusive)
  long generate() 
  { 
    return RANDOM_INTEGER % _range + _minimum; 
  }

  double mean() 
  {
    return _range / 2.0 + _minimum;
  }

  //

  Random_Uniform_Integer(long maximum=MAXLONG, long minimum=0)
  {
    set_parameter(maximum, minimum);
    initialize_generator();
  }
};

//

class Random_Uniform_Double 
// generate floating-point random numbers between 0.0 and max range
{
  double _maximum;

public:

  void set_parameter(double maximum=1.0)
  {
    _maximum = maximum;
  }

  // Will generate number from 0 to max (inclusive)
  double generate() 
  { 
    return (double) RANDOM_INTEGER / MAXLONG * _maximum; 
  }

  double mean() 
  {
    return _maximum / 2.0;
  }

  //

  Random_Uniform_Double(double maximum=1.0)
  { 
    set_parameter(maximum);
    initialize_generator();
  }
};

//

class Random_Exponential
// generate floating-point exponetial random numbers with mean 1/rate
{
  double _rate;

public:

  void set_parameter(double rate=1.0)
  {
    _rate = rate;
  }

  // Will generate number from 0 to max (inclusive)
  double generate() 
  { 
    return -log((double) RANDOM_INTEGER / MAXLONG) / _rate;
  }

  double mean() 
  {
    return 1.0 / _rate;
  }

  //

  Random_Exponential(double rate=1.0)
  {
    set_parameter(rate);
    initialize_generator();
  }
};

#endif
