//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: relation.cc,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/relation.cc,v $
//  
//  File	      : relation.cc
//  Author(s)	      : Jaeho Lee <jaeho@schroedinger.engin.umich.edu>
//  
//  Description	      : Relation
//  
//  Original author(s): Jaeho Lee <jaeho@schroedinger.engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Mon Feb 14 10:41:18 1994
//  Last Modified By  : Jaeho Lee <jaeho@splat.engin.umich.edu>
//  Last Modified On  : Mon Feb 26 22:50:07 1996
//  Update Count      : 174
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>
#include "binding.h"
#include "exp.h"
#include "symtab.h"
#include "wm.h"
#include "relation.h"

//
// Static Class Members
//

SymbolTable* Relation::_relation_table = new SymbolTable(512, 128, 253);

//
//
//

Relation::Relation(Relation* r, Binding* binding)
{
  _id = r->_id;
  _arity = r->_arity;
  _args = explist_eval_new(r->_args, binding);
}

Relation* Relation::eval_args(Binding* binding)
{
  explist_eval(_args, binding);
  return this;
}


void Relation::format(ostream& s, Binding* b)
{
  s << get_name() << " ";
  explist_format(get_args(), s, b);
}

void Relation::print(ostream& s, Binding* b)
{
  format(s, b);
  s << "\n";
}

//
// match
//

int unify(const Relation* dst_relation, Binding* dst_binding,
	  const Relation* src_relation, Binding* src_binding)
{
// If the source & the destination relation does not match, return 0.
// Otherwise change the destination binding with linked variables to
// the source relation binding and return 1

  if (src_relation->get_id() != dst_relation->get_id())
    return 0;

  if (src_relation->get_arity() <= 0 && dst_relation->get_arity() <= 0) 
    return 1;

  ExpListIterator next_src_arg(src_relation->get_args());
  ExpListIterator next_dst_arg(dst_relation->get_args());

  Expression* 	src_arg; 
  Expression* 	dst_arg;

  Binding 	dst_binding_copy(dst_binding);
  
  while ((src_arg = next_src_arg()) && (dst_arg = next_dst_arg())) {
    Value src_val = src_arg->eval(src_binding);
    Value dst_val = dst_arg->eval(dst_binding);

    if (src_val.is_defined() && dst_val.is_defined())
      if (src_val == dst_val)
	continue;
      else
	return 0;

    if (dst_arg->is_variable()) {
      if (src_arg->is_variable())
	dst_binding_copy.link_variables(dst_arg, src_arg, src_binding);
      else
	dst_binding_copy.set_value(dst_arg, src_val);
    }
    else {
      if (src_val != dst_val)
	return 0;
    }
  }

  if (dst_binding) 
    *dst_binding = dst_binding_copy;

  return 1;
}

