//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: relation.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/relation.h,v $
//  
//  File	      : relation.h
//  Author(s)	      : Jaeho Lee <jaeho@schroedinger.engin.umich.edu>
//  
//  Description	      : Relations
//  
//  Original author(s): Jaeho Lee <jaeho@schroedinger.engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Mon Feb 14 10:39:46 1994
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:06:58 1996
//  Update Count      : 85
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __RELATION_H__
#define __RELATION_H__

#include <string.h>
#include "exp.h"
#include "symtab.h"

class Binding;
class WmRelation;
class WmTable;

//
// RELATION
//

class Relation
{
  static SymbolTable*	_relation_table;
  long			_id;
  ExpList*		_args;
  long			_arity;

  //
  
  void init(const char* s, ExpList* el=0)
  {
    if ((_id=_relation_table->get_id(s)) < 0)
      _id = _relation_table->add(new Symbol(s))->get_id();
    _args = el ? el : new ExpList;
    _arity = _args->get_count();
  }
  
public:

  ~Relation()
  {
    _args->delete_entities();
    delete _args;
  }

  Relation(const char* s)
  {
    init(s);
  }
  
  Relation(const char* s, ExpList* el)
  {
    init(s, el);
  }
  
  Relation(Relation* r, Binding* binding=0);
  
  // access functions
  const long get_id() const
  {
    return _id;
  }
  
  const char* get_name() const
  {
    return _relation_table->lookup(_id)->get_name();
  }
  
  long get_arity() const
  {
    return _arity;
  }
  
  ExpList* get_args() const
  {
    return _args;
  }

  Relation* eval_args(Binding* b);

public:
  
  void print(ostream& s = cout, Binding* b=0);
  void format(ostream& s, Binding* b=0);
};

int unify(const Relation* dst_relation, Binding* dst_binding,
	  const Relation* src_relation, Binding* src_binding);

#endif
