//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: run.cc,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/run.cc,v $
//  
//  File              : run.cc
//  Author(s)         : Marcus J. Huber <marcush@pelican.eecs.umich.edu>
//  
//  Description       : KA execution runtime information
//  
//  Original author(s): Marcus J. Huber <marcush@pelican.eecs.umich.edu>
//  Organization      : University of Michigan AI Lab
//  Created On        : Tue Aug 16 14:25:13 1994
//  Last Modified By  : Marc Huber <marcush@darwin>
//  Last Modified On  : Fri Apr 19 18:14:20 1996
//  Update Count      : 189
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
//
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>
#include "extern.h"
#include "ka.h"
#include "run.h"
#include "goal.h"
#include "is.h"
#include "action.h"

//
//
// KaRuntimeFrame
//
//

/////////////////////////////////////////////////////////////////////////////
//
// This should never actually get executed, but just to make sure.
//
KaBodyElement::KaReturnType KaRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  return KaBodyElement::KA_ELEMENT_FAILED;
}

//
//
// KaSimpleRuntimeFrame
//
//


/////////////////////////////////////////////////////////////////////////////
//
// Perform primitive function
//
KaBodyElement::KaReturnType KaSimpleRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  Action* action = ((KaBodySimpleElement *)get_this_element())->get_action();

  Action::ActStatus return_val = action->execute(b);
  
  if (return_val == Action::ACT_FAILED)
    return KaBodyElement::KA_ELEMENT_FAILED;
  else
    return KaBodyElement::KA_ELEMENT_COMPLETE;
}


//
//
// KaGoalRuntimeFrame
//
//

/////////////////////////////////////////////////////////////////////////////
//
// Add the subgoal to the bottom of the intention stack.  We also need to
// check if we're supposed to run through the normal or the EFFECTS section
// to set up correctly for execution.
//
void KaGoalRuntimeFrame::intend(SoakElement* s)
{
  SoakElement*	new_se = new SoakElement;

  // If new goal, then set pointers
  if (subgoal == 0)
    subgoal = s->_from_goal;
  
  // Need to copy SoakElement s
  *new_se = *s;
  
  // add instantiated KA to goal
  subgoal->set_intention(new_se);
  if (program.sim_mode) {
    if (new_se->_ka->get_effect())
      subgoal->set_runtime_frame(new_se->_ka->get_effect()->get_start_element()->new_runtime_frame());
    else
      subgoal->set_runtime_frame(0);
  }
  else
    subgoal->set_runtime_frame(new_se->_ka->get_body()->get_start_element()->new_runtime_frame());

  if (new_se->_from_goal->get_prev_goal())
    subgoal->set_prev_goal(new_se->_from_goal->get_prev_goal());
}


/////////////////////////////////////////////////////////////////////////////
//
// Keep track of subgoal success and failure.  Pretty much just a modified
// version of the intention structure's 'activate' function.
//
KaBodyElement::KaReturnType KaGoalRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  Goal*				new_goal;
  KaRuntimeFrame*		active_frame;
  KaBodyElement::KaReturnType	return_val;
  KaBodyAtomicElement*		failure_section;

  // First pass at detecting executing this within a CYCLE body.
  if (this_goal == 0) {
    cerr << "\nWarning!  Detected goal action within execution of CYCLE body.  Ignoring.\n\n";
    return KaBodyElement::KA_ELEMENT_COMPLETE;
  }
  
  // Check to see if we've already posted a subgoal.  If there is,
  // we then need to see if an intention has been found for it.  If
  // so, we can execute something in the intention.
  if (subgoal != 0) {

    if (subgoal->get_intention() != 0) {
      active_frame = subgoal->get_runtime_frame();
      if (active_frame != 0) {
	
	// At this point we'll be executing an action from a KA so we have
	// to check the subgoal to see if it's still valid.
	if (!subgoal->get_intention()->_ka->confirm_context(subgoal->get_intention_binding()) ||
	    (subgoal->get_status() == IS_ABANDONED)) {
	  
	  // OOPS!  subgoal invalid!
	  // Clean up the stack. Replace some of these with something like
	  // IS->force_goal_failure()???
	  subgoal->remove_intention(1);
	  program.goal_list->remove(subgoal);
	  subgoal->set_status(IS_FAILURE);
	  program.is->set_current_goal(this_goal);
	  return KaBodyElement::KA_ELEMENT_FAILED;
	}
	
	// NOW we can run something!
	return_val = active_frame->execute(subgoal->get_intention_binding(), subgoal);
	
	if (return_val == KaBodyElement::KA_ELEMENT_FAILED) {
	  
	  if (program.show_is || program.show_gl)
	    cout << "Subgoal " << subgoal->get_name() << " failed!\n";
	  
	  // Clean up the stack.  This might be the exact same as the context
	  // failure case above so we might replace this with
	  // IS->force_goal_failure()....
	  subgoal->remove_intention(1);
	  program.goal_list->remove(subgoal);
	  this_goal->set_subgoal(0);
	  subgoal = 0;
	  program.is->set_current_goal(this_goal);
	  return KaBodyElement::KA_ELEMENT_FAILED;
	}
	else if (return_val == KaBodyElement::KA_ELEMENT_INCOMP) {
	  return KaBodyElement::KA_ELEMENT_INCOMP;
      }
	else { // if (return_val == KaBodyElement::KA_ELEMENT_COMPLETE)
	  if (active_frame->get_this_element()->is_final_element()) {
	    subgoal->set_runtime_frame(0);
	  }
	  else {
	    subgoal->set_runtime_frame(active_frame->get_this_element()->get_next_element()->new_runtime_frame());
	  }
	  return KaBodyElement::KA_ELEMENT_INCOMP;
	}
      }
      
      else {
	subgoal->remove_intention(0);
	program.goal_list->remove(subgoal);
	subgoal = 0;
	this_goal->set_subgoal(0);
	program.is->set_current_goal(this_goal);
	return KaBodyElement::KA_ELEMENT_COMPLETE;
      }
    }
    else { // no intention for subgoal yet
      this_goal->set_status(IS_BLOCKED);
      return KaBodyElement::KA_ELEMENT_INCOMP;
    }
  }

  // Otherwise we need to post the subgoal to the system.
  else {
    new_goal = program.goal_list->add((GoalAction*)((KaBodySimpleElement *)get_this_element())->get_action(),
				      this_goal->get_current_goal());

    // Set Frame's subgoal field
    subgoal = new_goal;
    subgoal->set_status(IS_ACTIVE);
    
    if (program.show_gl) {
      cout << "Just posted subgoal " << new_goal->get_name() << " to system.\n";
      program.goal_list->print();
    }
    
    return KaBodyElement::KA_ELEMENT_INCOMP;
  }
}


//
//
// KaBranchRuntimeFrame
//
//

inline KaBodyElement* KaBranchRuntimeFrame::get_branch_current(long branchnum) const
{
  if (branchnum >= 0 &&
      branchnum < ((KaBodyBranchElement *)get_this_element())->get_num_branches())
    return branch_current[branchnum];
  else
    return 0;
}


inline void KaBranchRuntimeFrame::set_branch_current(long branchnum, KaBodyElement* re)
{
  branch_current[branchnum] = re;
}


inline void KaBranchRuntimeFrame::set_active_branch_num(long n)
{
  if (n >= 0 && n < ((KaBodyBranchElement *)get_this_element())->get_num_branches()) 
    active_branch_num = n;
  else
    active_branch_num = -1;
}


/////////////////////////////////////////////////////////////////////////////
//
// Keep track of which branch we're executing, and whether we've succeeded
// or failed based on the branch type.
//
KaBodyElement::KaReturnType KaBranchRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  KaBodyElement*		curr_el;
  KaBodyElement::KaReturnType	return_val;

  curr_el = branch_current[active_branch_num];
  
  if (subframe == 0) {
    set_subframe(curr_el->new_runtime_frame());
  }
  
  return_val = subframe->execute(b, this_goal);
  
  // Something in the branch failed so this element may fail
  if (return_val == KaBodyElement::KA_ELEMENT_FAILED) {
   
    // If this is an AND branch, or if we've gone through all branches of an OR,
    // then the element has failed.
    if ((((KaBodyBranchElement *)get_this_element())->get_branch_type() == AND_BRANCH) ||
	(active_branch_num ==
	 ((KaBodyBranchElement *)get_this_element())->get_num_branches() - 1)) {
      return KaBodyElement::KA_ELEMENT_FAILED;
    }
    else {
      set_active_branch_num(active_branch_num + 1);
      branch_current[active_branch_num] =
	((KaBodyBranchElement *)this_element)->get_branch_start(active_branch_num);

      set_subframe(branch_current[active_branch_num]->new_runtime_frame());
      return KaBodyElement::KA_ELEMENT_INCOMP;
    }
    
    // The subframe failed, so get rid of it
    set_subframe(0);
  }
    
  // Nothing's been determined at this point
  else if (return_val == KaBodyElement::KA_ELEMENT_INCOMP) {
    return KaBodyElement::KA_ELEMENT_INCOMP;
  }

  else { // return_val == KaBodyElement::KA_ELEMENT_COMPLETE

    // Check to see if we can step to the next element
    if (!curr_el->is_final_element()) {
      branch_current[active_branch_num] = 
	branch_current[active_branch_num] = curr_el->get_next_element();

      set_subframe(branch_current[active_branch_num]->new_runtime_frame());
      return KaBodyElement::KA_ELEMENT_INCOMP;
    }
    else {
      // If this is an OR branch, or if we've gone through all branches of an AND,
      // then the element has succeeded.
      if ((((KaBodyBranchElement *)get_this_element())->get_branch_type() == OR_BRANCH) ||
	  (active_branch_num ==
	   ((KaBodyBranchElement *)get_this_element())->get_num_branches() - 1)) {
	return KaBodyElement::KA_ELEMENT_COMPLETE;
      }
      else { // Need to keep trucking
	set_active_branch_num(active_branch_num + 1);
	branch_current[active_branch_num] =
	  ((KaBodyBranchElement *)this_element)->get_branch_start(active_branch_num);
	set_subframe(branch_current[active_branch_num]->new_runtime_frame());
	return KaBodyElement::KA_ELEMENT_INCOMP;
      }
    }    
  }
}


//
//
// KaWhenRuntimeFrame
//
//

/////////////////////////////////////////////////////////////////////////////
//
// Check the test and step through the body if it's true.
//
KaBodyElement::KaReturnType KaWhenRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  Action::ActStatus		act_return_val;
  KaBodyElement::KaReturnType	body_return_val;

  // If we're at the beginning of the action, then we need to evaluate
  // the test.  If the test fails, we're simply done.  If it succeeds, we
  // go on and execute the first action.
  if (current_element == ((KaBodyWhenElement *)get_this_element())->get_start_element()) {

    act_return_val = ((KaBodyWhenElement *)get_this_element())->get_test()->execute(b);
   
    if (act_return_val != Action::ACT_SUCCEEDED) {
     
      // The subframe's done, so get rid of it
      set_subframe(0);
      return KaBodyElement::KA_ELEMENT_COMPLETE;
    }
  }

  if (subframe == 0) {
    set_subframe(current_element->new_runtime_frame());
  }
    
  body_return_val = subframe->execute(b, this_goal);
    
  if (body_return_val == KaBodyElement::KA_ELEMENT_FAILED) {
      
    // The subframe failed, so get rid of it
    set_subframe(0);
    return KaBodyElement::KA_ELEMENT_FAILED;
  }
  else if (body_return_val == KaBodyElement::KA_ELEMENT_COMPLETE) {

    // Check to see if the action is finished
    if (current_element->is_final_element()) {

      // The subframe's done, so first get rid of it
      set_subframe(0);
      return KaBodyElement::KA_ELEMENT_COMPLETE;
    }
    else {

      // Not done yet, so go on to the next action
      set_subframe(current_element->get_next_element()->new_runtime_frame());
      current_element = current_element->get_next_element();
    }
    
    return KaBodyElement::KA_ELEMENT_INCOMP;
  }
  else { // return_val == KaBodyElement::KA_ELEMENT_INCOMP
    return KaBodyElement::KA_ELEMENT_INCOMP;
  }
}


//
//
// KaWhileRuntimeFrame
//
//

/////////////////////////////////////////////////////////////////////////////
//
// Check the test and step through the body if it's true.  When the body is
// done, go to the top and check the test again, etc.
//
KaBodyElement::KaReturnType KaWhileRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  Action::ActStatus		act_return_val;
  KaBodyElement::KaReturnType	body_return_val;

  // If we're at the top of the loop we evaluate the test and execute the
  // first step if it succeeds.  Otherwise the loop is done.
  if (check_loop_cond) {

    act_return_val = ((KaBodyWhileElement *)get_this_element())->get_test()->execute(b);
   
    if (act_return_val != Action::ACT_SUCCEEDED) {
     
      // The subframe's done, so get rid of it
      set_subframe(0);
      return KaBodyElement::KA_ELEMENT_COMPLETE;
    }

    check_loop_cond = 0;
  }

  if (subframe == 0) {
    set_subframe(current_element->new_runtime_frame());
  }
    
  body_return_val = subframe->execute(b, this_goal);
    
  if (body_return_val == KaBodyElement::KA_ELEMENT_FAILED) {
      
    // The subframe failed, so get rid of it
    set_subframe(0);
    return KaBodyElement::KA_ELEMENT_FAILED;
  }
  else if (body_return_val == KaBodyElement::KA_ELEMENT_COMPLETE) {
    if (!current_element->is_final_element()) {
      set_subframe(current_element->get_next_element()->new_runtime_frame());
      current_element = current_element->get_next_element();
    }
    else {
      current_element =
	((KaBodyWhileElement *)get_this_element())->get_start_element();
      check_loop_cond = 1;
      set_subframe(0);
    }
    
    return KaBodyElement::KA_ELEMENT_INCOMP;
  }
  else { // return_val == KaBodyElement::KA_ELEMENT_INCOMP
    return KaBodyElement::KA_ELEMENT_INCOMP;
  }
}


//
//
// KaDoRuntimeFrame
//
//

/////////////////////////////////////////////////////////////////////////////
//
// Step through the body and THEN check the test.  If true, go back to the
// top, do the body again, etc.
//
KaBodyElement::KaReturnType KaDoRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  Action::ActStatus		act_return_val;
  KaBodyElement::KaReturnType	body_return_val;

  // Execute the current KA element in the loop
 if (current_element != 0) {

   body_return_val = subframe->execute(b, this_goal);

   if (body_return_val == KaBodyElement::KA_ELEMENT_FAILED) {
     set_subframe(0);
     return KaBodyElement::KA_ELEMENT_FAILED;
   }
   else if (body_return_val == KaBodyElement::KA_ELEMENT_COMPLETE) {
     current_element = current_element->get_next_element();
     if (current_element == 0)
       set_subframe(0);
     else {
       set_subframe(current_element->new_runtime_frame());
     }
     return KaBodyElement::KA_ELEMENT_INCOMP;
   }
   else { // return_val == KaBodyElement::KA_ELEMENT_INCOMP) {
     return KaBodyElement::KA_ELEMENT_INCOMP;
   }
 }

 else {
   // If we're at the bottom of the loop we evaluate the test and go back
   // up to the first step if it succeeds.  Otherwise the loop is done.
   act_return_val = ((KaBodyWhileElement *)get_this_element())->get_test()->execute(b);
   if (act_return_val != Action::ACT_SUCCEEDED) {
     return KaBodyElement::KA_ELEMENT_COMPLETE;
   }
   else {
     current_element = ((KaBodyDoElement *)get_this_element())->get_start_element();

     set_subframe(current_element->new_runtime_frame());
     return KaBodyElement::KA_ELEMENT_INCOMP;
   }
 }
}


//
//
// KaAtomicRuntimeFrame
//
//

/////////////////////////////////////////////////////////////////////////////
//
// Go through the entire body without returning back to the interpreter.
//
KaBodyElement::KaReturnType KaAtomicRuntimeFrame::execute(Binding* b, Goal* this_goal)
{
  Action::ActStatus		act_return_val;
  KaBodyElement::KaReturnType	body_return_val;
  KaBodyElement*		current;

  current = ((KaBodyAtomicElement *)get_this_element())->get_start_element();
  
  if (subframe == 0) {
    set_subframe(current->new_runtime_frame());
  }

  while (current) {

    body_return_val = subframe->execute(b, this_goal);
    if (body_return_val == KaBodyElement::KA_ELEMENT_FAILED) {
      
      // The subframe failed, so get rid of it
      set_subframe(0);
      return KaBodyElement::KA_ELEMENT_FAILED;
    }
    
    else if (body_return_val == KaBodyElement::KA_ELEMENT_COMPLETE) {
      if (!current->is_final_element()) {
	set_subframe(current->get_next_element()->new_runtime_frame());
      }
    }
    else { // if (body_return_val == KaBodyElement::KA_ELEMENT_INCOMP)
      continue;
    }
    current = current->get_next_element();
  }

  // Finished successfully
  return KaBodyElement::KA_ELEMENT_COMPLETE;
}


