//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: soak.h,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/soak.h,v $
//  
//  File	      : soak.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//
//  Description	      : Set of Applicable Knowledge-area
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:20:40 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:03:31 1996
//  Update Count      : 144
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __SOAK_H__
#define __SOAK_H__

#include "binding.h"

// Classes
class ostream;

class Ka;
class Goal;
class Binding;
class SoakContext;

class KaTable;
class WmTable;
class GoalSet;


//

class SoakElement
{ 

public:

  Ka*		_ka; 
  Goal*		_from_goal;
  Binding*	_binding;

  ~SoakElement()
  {
    delete _binding;
  }
  
  SoakElement()
  {
    _ka = 0;
    _from_goal = 0;
    _binding = 0;
  }

  void operator=(SoakElement& se)
  {
    if (this == &se) return;
    _ka = se._ka;
    _from_goal = se._from_goal;
    delete _binding;
    _binding = new Binding(se._binding);
  }

  double eval_priority();

  void print(ostream& s=cout);
};


class Soak
{
protected:
  Dlist<SoakElement*>*	_elements;

  SoakElement*	add(Ka* ka, Goal* g, Binding* b);
  inline long	random_uniform(long range) const;
  
public:

  Soak()
  {
    _elements = new Dlist<SoakElement*>;
  }
  
  ~Soak()
  {
    _elements->delete_entities();
    delete _elements;
  }

  Soak(KaTable* kt, WmTable* wt, GoalSet* tl, int test_generate=0);

  long		get_size() const	{ return _elements->get_count(); }

  SoakElement*	get_se(long nth);

  Dlist<SoakElement*>*	get_elements()	{ return _elements; }

  SoakElement*	get_first()		{ return _elements->first(); }

  SoakElement*	get_random();

  SoakElement*	get_priority_first();

  SoakElement* 	get_priority_random();
       
  void 		instantiate(Ka* ka, Binding* ka_binding, Goal* goal);

  void		operator=(Soak& s);

  void		print(ostream& s=cout);
};

//
//
//

typedef DlistIterator<SoakElement*>	SoakIterator;

//
// Soak List
//

class SoakList
{
  Dlist<Soak*>	*soaks;
  
public:

  ~SoakList()
  {
    soaks->delete_entities();
    delete soaks;
  }
  
  SoakList()
  {
    soaks = new Dlist<Soak*>;
  }

  Soak* get_soak(long level);

  Soak* add(Soak *soak);

  long size() const		{ return soaks->get_count(); }
};

#endif
