//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: symtab.cc,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/symtab.cc,v $
//  
//  File	      : symtab.cc
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : symbol table
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:19:57 1993
//  Last Modified By  : Jaeho Lee <jaeho@splat.engin.umich.edu>
//  Last Modified On  : Mon Feb 26 23:16:15 1996
//  Update Count      : 147
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>
#include <string.h>

#include "symtab.h"

//
// Static Member
// 
const long SymbolTable::_NullId = -1;

//
// Auxiliary Functions
//
int symbol_pointer_equal(Symbol* a, Symbol* b)
{
  return (a == b);
}

//
//
//

Symbol* SymbolTable::lookup(const long id) const
{
  return (id < 0 || id > _symbtab_size) ? 0 : _symbtab[id];
}

Symbol* SymbolTable::lookup(const char* name) const
{
  SymbolListIterator next(get_bucket(name));
  Symbol* s;

  while (s=next())
    if (strcmp(name, s->get_name()) == 0) return s;

  return 0; 
}

Symbol* SymbolTable::lookup(Symbol* symbol) const
{
  SymbolListIterator next_match(get_bucket(symbol));

  return next_match();
}

Symbol* SymbolTable::add(Symbol* symbol)
{
  // First see if the symbol table has any room left.
  if (_nextid >= _symbtab_size) {
    Symbol** new_symbtab = new Symbol*[_symbtab_size + _symbtab_inc_size];

    for (long i = 0; i < _symbtab_size; i++) 
      new_symbtab[i] = _symbtab[i];

    delete[] _symbtab;

    _symbtab = new_symbtab;
    _symbtab_size += _symbtab_inc_size;
  }
  
  // Put new entry in symbol table.
  symbol->set_id(_nextid);
  _symbtab[_nextid++] = symbol;
  
  // Now install symbol in the hash table. Put new one at head of its hash bucket
  _hashtab[hash(symbol->get_name())].insert(symbol);

  return symbol;
}


void SymbolTable::remove_this(SymbolListIterator* bucket)
{
  Symbol* symbol = bucket->get_this();

  // remove from the hashtable bucket
  bucket->remove_this();

  // remove from the symbol table
  _symbtab[--_nextid]->set_id(symbol->get_id());
  _symbtab[symbol->get_id()] = _symbtab[_nextid];
  delete symbol;
}

     
void SymbolTable::remove(Symbol* symbol)
{
  SymbolListIterator bucket(get_bucket(symbol));
  Symbol* symbol_found = bucket(symbol_pointer_equal, symbol);

  if (!symbol_found) return;

  remove_this(&bucket);
}


Symbol* SymbolTable::replace(Symbol* old_s, Symbol* new_s)
{
// replacement is only applicable to the symbol with the same symbol name
// because it should not change the hash table

  long old_hash = hash(old_s->get_name());
  long new_hash = hash(new_s->get_name());

  if (old_hash != new_hash) return 0;

  SymbolListIterator bucket(get_bucket(old_s));
  Symbol* symbol_found = bucket(symbol_pointer_equal, old_s);

  if (!symbol_found) return 0;

  // replace from the hashtable bucket
  bucket.replace_this(new_s);

  // replace from the symbol table
  new_s->set_id(old_s->get_id());
  _symbtab[old_s->get_id()] = new_s;
  delete old_s;

  return new_s;
}

