//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: symtab.h,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/symtab.h,v $
//  
//  File	      : symtab.h
//  Author(s)	      : Jaeho Lee <jaeho@engin.umich.edu>
//  
//  Description	      : symbol table (variable table)
//  
//  Original author(s): Jaeho Lee <jaeho@engin.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 12:19:27 1993
//  Last Modified By  : Jaeho Lee <jaeho@splat.engin.umich.edu>
//  Last Modified On  : Mon Feb 26 23:53:38 1996
//  Update Count      : 242
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __SYMTAB_H__
#define __SYMTAB_H__

//
#include <stream.h>
#include <string.h>

#include "dlist.h"

//

#define SYMBTAB_SIZE		16
#define HASHTAB_SIZE		7  
#define SYMBTAB_INCREMENT	16

//
// Symbol
//
class Symbol
{
friend class SymbolTable;

private:
  
  char*	_name;
  long	_id;
	
  //

  void init(const char* name, const long id=-1)
  {
    if (name)
      ::strcpy(_name=new char[::strlen(name)+1], name);
    else
      _name = 0;
    
    _id = id;
  }

public:
  
  virtual ~Symbol()
  {
    delete[] _name;
  }

  Symbol(const char* name="", long id=-1)
  {
    init(name, id);
  }

  Symbol(const Symbol& s)
  {
    init(s._name, s._id);
  }

public:
  
  const char*	get_name() const	{ return _name; }

  const long	get_id() const		{ return _id; }
  void		set_id(long sid)	{ _id = sid; }

  // 

  virtual void print(ostream& s=cout) const
  {
    s << _id << ": " << _name;
  }
};


//
// Dlist
//

typedef Dlist<Symbol*>			SymbolList;
typedef DlistIterator<Symbol*>		SymbolListIterator;

//
// 
//

class SymbolTable
{
  Symbol**	_symbtab;
  SymbolList*	_hashtab;

private:

  static const long _NullId;
  
protected:

  long		_symbtab_size;
  long		_hashtab_size;

  long		_nextid;
  long		_symbtab_inc_size;

protected:
  
  long hash(const char* name) const
  {
    register long key = 0;

    while (*name != '\0')
      key += *name++;

    return (key % _hashtab_size); 
  }
  
public:

  virtual ~SymbolTable()
  {
    long i;
    
    for (i = 0; i < get_size(); i++)
      delete _symbtab[i];
    delete [] _symbtab;

    delete [] _hashtab;
  }
  
  SymbolTable(long symbtab_sz		= SYMBTAB_SIZE, 
	      long symbtab_inc_sz	= SYMBTAB_INCREMENT,
	      long hashtab_sz		= HASHTAB_SIZE)
  {
    _symbtab = new Symbol*[_symbtab_size = symbtab_sz];
    _hashtab = new SymbolList[_hashtab_size = hashtab_sz];
    _symbtab_inc_size = symbtab_inc_sz;
    _nextid = 0;
  }

public:
  
  Symbol*	lookup(const long id) const;
  Symbol*	lookup(const char* name) const;
  Symbol*	lookup(Symbol* symbol) const;

  Symbol*	add(Symbol* symbol);

  void		remove_this(SymbolListIterator* bucket);
  void		remove(Symbol* symbol);

  Symbol*	replace(Symbol* old_s, Symbol* new_s);


public:

  long get_size() const
  {
    return _nextid;
  }

  SymbolList& get_bucket(const char* name) const
  {
    return _hashtab[hash(name)];
  }
  
  SymbolList& get_bucket(Symbol* symbol) const
  {
    return get_bucket(symbol->get_name());
  }
  
  const long get_id(const char* name) const
  {
    Symbol* s = lookup(name);
    return s ? s->get_id() : _NullId;
  }

  //
  
  virtual void print(ostream& s=cout) const
  {
    for (long i = 0; i < _nextid; i++) {
      lookup(i)->print();
      s << "\n";
    }
  }
};

#endif
