//  -*- Mode: C++ -*-
//////////////////////////////////////////////////////////////////////////////
//  
//  $Id: wm.cc,v 3.1 1996/02/16 18:23:00 jaeho Exp $
//  $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/wm.cc,v $
//  
//  File	      : wm.cc
//  Author(s)	      : Jaeho Lee <jaeho@dworkin.eecs.umich.edu>
//  
//  Description	      : World Model 
//  
//  Original author(s): Jaeho Lee <jaeho@dworkin.eecs.umich.edu>
//  Organization      : Distributed Intelligent Agents Group (UM-DIAG)
//  Created On	      : Sat Jul	 3 15:33:33 1993
//  Last Modified By  : Jaeho Lee <jaeho@cavs.engin.umich.edu>
//  Last Modified On  : Fri Jan 12 16:04:43 1996
//  Update Count      : 203
//  
//////////////////////////////////////////////////////////////////////////////
//
//  UM-PRS
//
//  Copyright 1996
//  The Regents of the University of Michigan
//  All Rights Reserved
//
//  This software is provided as is, without representation as to its
//  fitness for any purpose, and without warranty of any kind, either
//  express or implied, including without limitation the implied
//  warranties of merchantability and fitness for a particular purpose.
//  The Regents of the University of Michigan shall not be liable for any
//  damages, including special, indirect, incidental, or consequential
//  damages, with respect to any claim arising out of or in connection
//  with the use of the software, even if it has been or is hereafter
//  advised of the possibility of such damages.
// 
//	    This work has been sponsored in part by:
//    the Department of Defense Advanced Research Projects Agency
//		 under grant DAAE-07-92-C-R012.
//
//////////////////////////////////////////////////////////////////////////////

#include <stream.h>
#include <string.h>
#include "symtab.h"
#include "binding.h"
#include "relation.h"
#include "exp.h"
#include "wm.h"

//
//
//

int WmRelation::match_relation(const Relation* patt_relation, Binding* patt_binding)
{
  if (unify(patt_relation, patt_binding, get_relation(), 0)) {
    if (patt_binding)
      patt_binding->check_new_wm_binding(is_new());
    return 1;
  }
  else
    return 0;
}

//

int WmTable::match(Relation* relation, Binding* binding)
{
  WmTableBucketIterator next_wr(this, relation);

  return next_wr(binding) ? 1 : 0;
}

//
//
//

void WmTable::assert(Relation* r, Binding* b)
{
  if (!match(r, b))
    // note creating a new relation with args evaluated
    add(new WmRelation(new Relation(r, b)));
}


void WmTable::retract(Relation* r, Binding* b)
{
  WmTableBucketIterator next_wr(this, r);
  WmRelation* wr;
  
  while(wr = next_wr(b))
    next_wr.remove_this();
} 


void WmTable::update(Relation* old_rel, Relation* new_rel, Binding* b)
{
  retract(old_rel, b);
  assert(new_rel, b);
}

//

void WmTable::clear_new_all() 
{
  for (register long i = 0; i < _nextid; i++)
    lookup(i)->clear_new();
}

int WmTable::any_new() 
{
  for (register long i = 0; i < _nextid; i++) {
    if (lookup(i)->is_new()) {
      return 1;
    }
  }
  return 0;
}


void WmTable::print(ostream& s) const
{
  register Relation* r;
  register WmRelation* w;
  Expression* e;

  s << "\nWorld Model\n";
  for (register long i = 0; i < _nextid; i++) {
    w = lookup(i);
    r = w->get_relation();

    s << "[" << form("%1s] %4d:", w->is_new()? "N":" ", i);
    r->print(s);
  }
}
