/*
 * -*- Mode: C -*-
 ******************************************************************************
 *
 * $Id: yyparse.y,v 3.1.1.2 1996/02/27 06:48:39 jaeho Exp $
 * $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/yyparse.y,v $
 *
 * File		     : parse.y
 * Author(s)	     : Jaeho Lee <jaeho@engin.umich.edu>
 *
 * Description	     : parser
 *
 * Original author(s): Jaeho Lee <jaeho@dworkin.eecs.umich.edu>
 * Organization	     : Distributed Intelligent Agents Group (UM-DIAG)
 * Created On	     : Sat Jul	3 12:23:37 1993
 * Last Modified By  : Marcus J. Huber <marcush@ultra>
 * Last Modified On  : Fri Feb 21 12:57:59 1997
 * Update Count	     : 559
 *
 ******************************************************************************
 *
 *  UM-PRS
 *
 *  Copyright 1996
 *  The Regents of the University of Michigan
 *  All Rights Reserved
 *
 *  This software is provided as is, without representation as to its
 *  fitness for any purpose, and without warranty of any kind, either
 *  express or implied, including without limitation the implied
 *  warranties of merchantability and fitness for a particular purpose.
 *  The Regents of the University of Michigan shall not be liable for any
 *  damages, including special, indirect, incidental, or consequential
 *  damages, with respect to any claim arising out of or in connection
 *  with the use of the software, even if it has been or is hereafter
 *  advised of the possibility of such damages.
 *
 *	    This work has been sponsored in part by:
 *    the Department of Defense Advanced Research Projects Agency
 *		 under grant DAAE-07-92-C-R012.
 *
 ******************************************************************************
 */

%{
#define	YYDEBUG		1
#include <stdio.h>

#include "extern.h"
#include "exp.h"
#include "exp-val.h"
#include "exp-var.h"
#include "exp-fun.h"
#include "predicate.h"
#include "primitive.h"
#include "relation.h"
#include "condition.h"
#include "symtab.h"
#include "action.h"
#include "goal.h"
#include "wm.h"
#include "ka.h"


//
// Declarations
//
int  yyparse();
int  yylex();
void yyerror(char* s);

//
// Locals
//
Function		function		= 0;
Ka*			current_ka		= 0;
Action*			true_test		= new Test(new Value(1));
%}



%union {
  long			ival;
  double		fval;
  char			cval;
  char			sval[5000];
  Expression*		exp;
  ExpList*		expl;
  Relation*		rel;
  Condition*		con;
  ConditionList*	conl;
  Action*		action;
  GoalAction*		goal_action;
  KaBodyElement*	body_element;
  KaBodyElement*	body_elements;

  KaBodyBranchElement*	branch_element;
}

%token			GOAL_DECL
%token			FACT_DECL
%token			CYCLE_DECL

%token			KA_START
%token			KA_NAME
%token			KA_DOC
%token			KA_CONTEXT
%token			KA_PRIORITY
%token			KA_EFFECT
%token			KA_FAILURE

%token			KA_PURPOSE
%token			KA_BODY
%token			KA_BODY_AND
%token			KA_BODY_OR
%token			KA_BODY_DO
%token			KA_BODY_WHILE
%token			KA_BODY_WHEN
%token			KA_BODY_ATOMIC

%token			EXECUTE
%token			LOAD
%token			PARSE
%token			ASSIGN
%token			FAIL
%token			RETRIEVE
%token			TEST
%token			POST
%token			UNPOST
%token			ACHIEVE
%token			MAINTAIN
%token			QUERY
%token			ASSERT
%token			RETRACT
%token			UPDATE

%token			KEYWORD_TEST
%token			KEYWORD_PRIORITY
%token			KEYWORD_BY
%token			KEYWORD_NOT_BY

%token	<ival>		INTEGER
%token	<fval>		FLOAT
%token	<sval>		STRING
%token	<ival>		ADDRESS

%token	<sval>		IDENTIFIER
%token	<sval>		VARIABLE

%token			'('
%token			')'
%token			'{'
%token			'}'

%token	<sval>		'+'	      /* plus */
%token	<sval>		'-'	      /* minus */
%token	<sval>		'*'	      /* times */
%token	<sval>		'/'	      /* divide */
%token	<sval>		'%'	      /* modulo */

%token	<sval>		EQ
%token	<sval>		NEQ
%token	<sval>		LT
%token	<sval>		LE
%token	<sval>		GT
%token	<sval>		GE

%token	<sval>		AND
%token	<sval>		OR
%token	<sval>		NOT

%token	<sval>		ASSERT
%token	<sval>		FACT
%token	<sval>		RETRIEVE


%type	<sval>		function_name

%type	<exp>		value
%type	<exp>		variable
%type	<exp>		funcall
%type	<exp>		predicate
%type	<exp>		expression
%type	<expl>		explist
%type	<con>		condition
%type	<conl>		condition_list

%type	<action>	action
%type	<action>	exec_action
%type	<goal_action>	goal_action
%type	<action>	misc_action

%type	<body_element>		ka_body_element
%type	<body_element>		ka_body_elements
%type	<body_element>		ka_branch
%type	<branch_element>	ka_branches

%start	program

/* Grammar follows */

%%

program		: declaration_seq
		  {
		    parser.stop_if_errors();
		  }
;



declaration_seq	: /* empty */
		| declaration_seq declaration
;


declaration	: fact_decl
		| goal_decl
		| cycle_decl
		| ka
;

/***********/
/* Assert  */
/* Task	   */
/***********/

fact_decl	: FACT_DECL facts
;

facts		: /* empty */
		| facts fact
;

fact		: IDENTIFIER explist ';'
		  {
		    // program.world_model->assert(new Relation($1, $2));
		    Relation r($1, $2);
		    program.world_model->assert(&r);
		  }
;

goal_decl	: GOAL_DECL goals
;

goals		: /* empty */
		| goals goal_action ';'
		  {
		    $2->set_trace(parser.source_file_name, parser.source_line_number);
		    program.goal_list->add_unique($2);
		  }
;


cycle_decl	: CYCLE_DECL
		  {
		    current_ka = new Ka;
		  }
		'{' ka_body_elements '}'
		  { 
		    current_ka->set_body(new KaBody($4));
		    program.cycle = current_ka;
		  }
;


/**********************************************************************/
/*								      */
/*	KA							      */
/*								      */
/**********************************************************************/

ka		: error '}'		
		  {
		    parser.stop_on_too_many_errors();
		    yyerrok; 
		  }
		| KA_START
		  {
		    current_ka = new Ka;
		  }
		  '{' ka_components '}'
		  { 
		    program.ka_library->add(new KaEnt(current_ka));
		  }
;

ka_components	: ka_component
		| ka_components ka_component
;

ka_component	: ka_name
		| ka_doc
		| ka_purpose
		| ka_context
		| ka_priority
		| ka_effect
		| ka_failure
		| ka_body
;


/***********/
/* Name	   */
/* Doc	   */
/***********/

ka_name		: KA_NAME STRING
		  {
		    current_ka->set_name($2);
		  }
;

ka_doc		: KA_DOC STRING
		  {
		    current_ka->set_doc($2);
		  }
;

/***********/
/* Purpose */
/***********/

ka_purpose	: KA_PURPOSE action ';'
		  {
		    current_ka->set_purpose($2);
		  }
;

/***********/
/* Context */
/***********/

ka_context	: KA_CONTEXT condition_list
		  {
		    current_ka->add_context($2);
		  }
;



/************/
/* Priority */
/************/

ka_priority	: KA_PRIORITY expression ';'
		  { 
		    current_ka->set_priority($2);
		  }
;

/*********************/
/* Effect / Failure  */
/*********************/

ka_effect	: KA_EFFECT ka_body_elements
		  { 
		    current_ka->set_effect(new KaBody($2));
		  }
;

ka_failure	: KA_FAILURE ka_body_elements
		  { 
		    current_ka->set_failure(new KaBodyAtomicElement($2));
		  }
;


/***********/
/* Body	   */
/***********/

ka_body		: KA_BODY ka_body_elements
		  { 
		    current_ka->set_body(new KaBody($2));
		  }
;

ka_branches	: ka_branch			
		  { 
		    $$ = new KaBodyBranchElement();
		    $$->add_branch($1);
		  }
		| ka_branches ka_branch
		  {
		    $1->add_branch($2);
		    $$ = $1;
		  }
;

ka_branch	: '{' ka_body_elements '}'		
		  {
		    $$ = $2;
		  }
;


ka_body_elements: ka_body_element
		  {
		    $$ = $1;
		  }
		| ka_body_element ka_body_elements
		  {
		    $1->set_next_element($2);
		    $$ = $1;
		  }

ka_body_element : action ';'
		  {
		    $$ = new KaBodySimpleElement($1);
		  }
		| KA_BODY_AND ka_branches ';'
		  { 
		    $2->set_branch_type(AND_BRANCH);
		    $$ = $2;
		  }
		| KA_BODY_OR  ka_branches ';'
		  {
		    $2->set_branch_type(OR_BRANCH);
		    $$ = $2;
		  }
		| KA_BODY_WHILE ':' action ka_branch ';'
		  {
		    $$ = new KaBodyWhileElement($3, $4);
		  }

		| KA_BODY_WHEN ':' action ka_branch ';'
		  {
		    $$ = new KaBodyWhenElement($3, $4);
		  }

		| KA_BODY_DO ka_branch KA_BODY_WHILE ':' action ';'
		  {
		    $$ = new KaBodyDoElement($5, $2);
		  }

		| KA_BODY_ATOMIC ka_branch ';'
		  {
		    $$ = new KaBodyAtomicElement($2);
		  }
;


/**********************************************************************/
/*								      */
/*	Condtion						      */
/*								      */
/**********************************************************************/

condition_list	: /* empty */
		  {
		    $$ = new ConditionList();
		  }
		| condition_list condition
		  {
		    $1->append($2); $$ = $1;
		  }
;

condition	: expression ';'	
		  {
		    $$ = new ExpCondition($1);
		  }
		| FACT IDENTIFIER explist ';'
		  {
		    $$ = new FactCondition(new Relation($2, $3));
		  }
		| RETRIEVE IDENTIFIER explist ';'
		  {
		    $$ = new RetrieveCondition(new Relation($2, $3));
		  }
;


/**********************************************************************/
/*								      */
/*	Action							      */
/*								      */
/**********************************************************************/

action		: exec_action
		  {
		    $$ = $1;
		    $$->set_trace(parser.source_file_name, parser.source_line_number);
		  }
		| misc_action
		  {
		    $$ = $1;
		    $$->set_trace(parser.source_file_name, parser.source_line_number);
		  }
		| goal_action
		  {
		    $$ = $1;
		    $$->set_trace(parser.source_file_name, parser.source_line_number);
		  }
;

exec_action	: EXECUTE function_name explist
		  {
		    if ((function = lookup_function_table($2)) == 0) {
		      yyerror(form("'%s' undefined function", $2));
		      parser.stop_on_too_many_errors();
		      $$ = new Primitive($2, function, $3);
		    }
		    else {
		      $$ = new Primitive($2, function, $3);
		    }
		  }

		| LOAD explist
		  {
		    $$ = new Load($2);
		  }

		| PARSE explist
		  {
		    $$ = new Parse($2);
		  }

		| RETRIEVE IDENTIFIER explist
		  {
		    $$ = new Retrieve(new Relation($2, $3));
		  }

		| TEST expression
		  {
		    $$ = new Test($2);
		  }

		| ASSERT IDENTIFIER explist
		  {
		    $$ = new Assert(new Relation($2, $3));
		  }

		| RETRACT IDENTIFIER explist
		  {
		    $$ = new Retract(new Relation($2, $3));
		  }

		| UPDATE '(' IDENTIFIER explist ')' '(' IDENTIFIER explist ')'
		  {
		    $$ = new Update(new Relation($3, $4), new Relation($7, $8));
		  }
;

misc_action	: ASSIGN variable expression
		  {
		    $$ = new Assign($2, $3);
		  }
		| FACT IDENTIFIER explist
		  {
		    $$ = new Fact(new Relation($2, $3));
		  }
		| FAIL
		  {
		    $$ = new Fail;
		  }
		| POST goal_action
		  {
		    $$ = new Post($2);
		  }
		| UNPOST goal_action
		  {
		    $$ = new UnPost($2);
		  }
;


goal_action	: ACHIEVE IDENTIFIER explist
		  {
		    $$ = new Achieve($2, new Relation($2, $3));
		  }
		| MAINTAIN IDENTIFIER explist
		  {
		    $$ = new Maintain($2, new Relation($2, $3));
		  }

		| QUERY	IDENTIFIER explist
		  {
		    $$ = new Query(new Relation($2, $3));
		  }

		| QUERY	IDENTIFIER explist KEYWORD_TEST expression
		  {
		    $$ = new Query(new Relation($2, $3), $5);
		  }
;

goal_action	: goal_action KEYWORD_PRIORITY expression
		  {
		    $1->set_priority($3);
		    $$ = $1;
		  }
;

goal_action	: goal_action KEYWORD_BY explist
		  {
		    $1->set_by($3);
		    $$ = $1;
		  }
;

goal_action	: goal_action KEYWORD_NOT_BY explist
		  {
		    $1->set_not_by($3);
		    $$ = $1;
		  }
;


/**********************************************************************/
/*								      */
/*	Expression						      */
/*								      */
/**********************************************************************/

expression	: error		
		  {
		    parser.stop_on_too_many_errors();
		    yyerrok; 
		  }
		| value		{ $$ = $1; }
		| variable	{ $$ = $1; }
		| funcall	{ $$ = $1; }
		| predicate	{ $$ = $1; }
;


explist		: /* empty */		
		  {
		    $$ = new ExpList();
		  }
		| explist expression	
		  {
		    $1->append($2); $$ = $1;
		  }
;


value		: INTEGER	{ $$ = new Value($1); }
		| FLOAT		{ $$ = new Value($1); }
		| STRING	{ $$ = new Value($1); }
		| ADDRESS	{ $$ = new Value((void*) $1); }
;

variable	: VARIABLE
		  {
		    $$ = new Variable(current_ka->get_symtab(), $1);
		  }
;


function_name	: IDENTIFIER
		| '+'	    
		| '-'	    
		| '*'	    
		| '/'	    
		| '%'	    
			    
		| EQ	    
		| NEQ	    
		| LT	    
		| LE	    
		| GT	    
		| GE	    
			    
		| AND	    
		| OR	    
		| NOT	    
;


funcall		: '(' function_name explist ')'
		  {
		    if ((function = lookup_function_table($2)) == 0) {
		      yyerror(form("'%s' undefined function", $2));
		      parser.stop_on_too_many_errors();
		      $$ = new Funcall($2, function, $3);
		    }
		    else {
		      $$ = new Funcall($2, function, $3);
		    }
		  }
;

predicate	: '(' FACT	IDENTIFIER explist ')'
		  {
		    $$ = new	PredicateFact($3, new Relation($3, $4));
		  }
		| '(' RETRIEVE	IDENTIFIER explist ')'
		  {
		    $$ = new	PredicateRetrieve($3, new Relation($3, $4));
		  }
		| '(' ACHIEVE	IDENTIFIER explist ')'
		  {
		    $$ = new	PredicateAchieve($3, new Relation($3, $4));
		  }
;

/*
id		: IDENTIFIER
		{
		  $$ = new Value($1);
		}
		| expression
		{
		  $$ = $1;
		}
;
*/

%%

#include <stream.h>

void yyerror (char* s)	/* Called by yyparse on error */
{
  cerr << "File: " << parser.source_file_name << ":error:";
  cerr << " line# " << parser.source_line_number;
  cerr << " (character# " << parser.source_char_position << ")";
  cerr << ":" << s << "\n";
}
