/* 
 * -*- Mode: C -*-
 ******************************************************************************
 * 
 * $Id: yytoken.l,v 3.1.1.1 1996/02/27 06:48:39 jaeho Exp $
 * $Source: /afs/engin.umich.edu/u/j/a/jaeho/UM-PRS/code3.1/RCS/yytoken.l,v $
 * 
 * File		     : token.l
 * Author(s)	     : Jaeho Lee <jaeho@dworkin.eecs.umich.edu>
 * 
 * Description	     : Lexical Analyzer
 * 
 * Original author(s): Jaeho Lee <jaeho@dworkin.eecs.umich.edu>
 * Organization	     : Distributed Intelligent Agents Group (UM-DIAG)
 * Created On	     : Sat Jul	3 12:18:46 1993
 * Last Modified By  : Jaeho Lee <jaeho@splat.engin.umich.edu>
 * Last Modified On  : Mon Feb 26 23:52:37 1996
 * Update Count	     : 88
 * 
 ******************************************************************************
 *
 *  UM-PRS
 *
 *  Copyright 1996
 *  The Regents of the University of Michigan
 *  All Rights Reserved
 *  
 *  This software is provided as is, without representation as to its
 *  fitness for any purpose, and without warranty of any kind, either
 *  express or implied, including without limitation the implied
 *  warranties of merchantability and fitness for a particular purpose.
 *  The Regents of the University of Michigan shall not be liable for any
 *  damages, including special, indirect, incidental, or consequential
 *  damages, with respect to any claim arising out of or in connection
 *  with the use of the software, even if it has been or is hereafter
 *  advised of the possibility of such damages.
 * 
 *	    This work has been sponsored in part by:
 *    the Department of Defense Advanced Research Projects Agency
 *		 under grant DAAE-07-92-C-R012.
 *
 ******************************************************************************
 */

%{
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "extern.h"
#include "exp.h"
#include "exp-val.h"
#include "exp-var.h"
#include "exp-fun.h"
#include "predicate.h"
#include "primitive.h"
#include "relation.h"
#include "condition.h"
#include "symtab.h"
#include "action.h"
#include "goal.h"
#include "wm.h"
#include "ka.h"

#include "yyparse.h"

//
//
//

/* yytext is too short for long comments such as those generated by RCS */
#undef	YYLMAX
#define YYLMAX	4096
#define YYEOF	0

//
//
//

extern int lex_getchar();

/* keep track of current character position */
#undef	YY_INPUT
#define YY_INPUT(buf,result,max_size)				\
	{							\
	  int c = lex_getchar();				\
	  parser.source_char_position++;			\
	  result = (c == EOF) ? YY_NULL : (buf[0]=c, 1);	\
	}

void scan_till(int);
static int readescape(int*);

%}

delim		[ \t\n]*
ws		[ \t\n]+
letter		[-_A-Za-z]
digit		[0-9]

variable	$({letter}|{digit}|".")*
identifier	[_A-Za-z]({letter}|{digit})*

string		\"([^\"\n]|\"\"|\\\")*\"
charac		('[^'\n]')|('''')

EXP		([eE][+-]?[0-9]+)
number0		[-+]?([0-9]+?)
number1		[-+]?([0-9]+"."[0-9]*{EXP}?)
number2		[-+]?("."[0-9]+{EXP}?)
hexa0		0x[0-9a-f]+

notnumber	[-a-zA-Z0-9\_\!\@\#\$\%\&\*\<\>\:\+\/]

%x COMMENT COMMENTCPP
%%

"/*"			{ BEGIN(COMMENT); }
"//"			{ BEGIN(COMMENTCPP); }

<COMMENT>[^*\n]*	/* eat anything that's not a '*' */
<COMMENT>"*"+[^*/\n]*	/* eat up '*'s not followd by '/'s */
<COMMENT>\n		{ parser.source_line_number++; }
<COMMENT>"*"+"/"	{ BEGIN(INITIAL); }

<COMMENTCPP>.		/* eat anything that's not a '\n' */
<COMMENTCPP>\n		{ parser.source_line_number++; BEGIN(INITIAL); }

<<EOF>>			{ yyterminate(); }
		
[ \t]+			;
[\n]			{ parser.source_line_number++; }
		
		
"{"			{ parser.level_section++;    return '{'; }
"}"			{ parser.level_section--;    return '}'; }
		
		
"("			{ parser.level_expression++; return '('; }
")"			{ parser.level_expression--; return ')'; }

"["			{ return '[';	}
"]"			{ return ']';	}
";"			{ return ';';	}
":"			{ return ':';	}
"&"			{ return '&';	}

"+"			{ strcpy(yylval.sval, yytext); return '+'; }
"-"			{ strcpy(yylval.sval, yytext); return '-'; }
"*"			{ strcpy(yylval.sval, yytext); return '*'; }
"/"			{ strcpy(yylval.sval, yytext); return '/'; }
"%"			{ strcpy(yylval.sval, yytext); return '%'; }
		
"=="			{ strcpy(yylval.sval, yytext); return EQ ; }
"!="			{ strcpy(yylval.sval, yytext); return NEQ; }
"<"			{ strcpy(yylval.sval, yytext); return LT ; }
"<="			{ strcpy(yylval.sval, yytext); return LE ; }
">"			{ strcpy(yylval.sval, yytext); return GT ; }
">="			{ strcpy(yylval.sval, yytext); return GE ; }
		
"&&"			{ strcpy(yylval.sval, yytext); return AND; }
"||"			{ strcpy(yylval.sval, yytext); return OR ; }
"!"			{ strcpy(yylval.sval, yytext); return NOT; }
		
FACT			{ strcpy(yylval.sval, yytext); return FACT; }



GOALS{delim}:		{ return GOAL_DECL; }
FACTS{delim}:		{ return FACT_DECL; }
CYCLE{delim}/"{"	{ return CYCLE_DECL; }

KA{delim}/"{"		{ return KA_START; }
NAME{delim}:		{ return KA_NAME; }
DOCUMENTATION{delim}:	{ return KA_DOC; }
PURPOSE{delim}:		{ return KA_PURPOSE; }
CONTEXT{delim}:		{ return KA_CONTEXT; }
PRIORITY{delim}:	{ return KA_PRIORITY; }
EFFECT{delim}:		{ return KA_EFFECT; }
FAILURE{delim}:		{ return KA_FAILURE; }

BODY{delim}:		{ return KA_BODY; }
AND			{ return KA_BODY_AND; }
OR			{ return KA_BODY_OR; }
DO			{ return KA_BODY_DO; }
WHILE			{ return KA_BODY_WHILE; }
WHEN			{ return KA_BODY_WHEN; }
ATOMIC			{ return KA_BODY_ATOMIC; }

":TEST"			{ return KEYWORD_TEST; }
":PRIORITY"		{ return KEYWORD_PRIORITY; }
":BY"			{ return KEYWORD_BY; }
":NOT-BY"		{ return KEYWORD_NOT_BY; }


EXECUTE			{ strcpy(yylval.sval, yytext); return EXECUTE;	}
LOAD			{ strcpy(yylval.sval, yytext); return LOAD;	}
PARSE			{ strcpy(yylval.sval, yytext); return PARSE;	}
ASSIGN			{ strcpy(yylval.sval, yytext); return ASSIGN;	}
RETRIEVE		{ strcpy(yylval.sval, yytext); return RETRIEVE;	}
TEST			{ strcpy(yylval.sval, yytext); return TEST;	}
POST			{ strcpy(yylval.sval, yytext); return POST;	}
UNPOST			{ strcpy(yylval.sval, yytext); return UNPOST;	}
ACHIEVE			{ strcpy(yylval.sval, yytext); return ACHIEVE;	}
MAINTAIN		{ strcpy(yylval.sval, yytext); return MAINTAIN;	}
QUERY			{ strcpy(yylval.sval, yytext); return QUERY;	}
ASSERT			{ strcpy(yylval.sval, yytext); return ASSERT;	}
FAIL			{ strcpy(yylval.sval, yytext); return FAIL;	}
RETRACT			{ strcpy(yylval.sval, yytext); return RETRACT;	}
UPDATE			{ strcpy(yylval.sval, yytext); return UPDATE;	}



{number0}		{ sscanf(yytext, "%ld", &yylval.ival); return INTEGER;}

({number1}|{number2})	{ sscanf(yytext, "%lf", &yylval.fval); return FLOAT; }

{hexa0}			{ sscanf(yytext, "%li", &yylval.ival); return ADDRESS; }

\"			{ scan_till('\"'); return(STRING); }

{identifier}		{ strcpy(yylval.sval, yytext); return IDENTIFIER; }

{variable}		{ strcpy(yylval.sval, yytext); return VARIABLE; }


%%

void scan_till(int c)
{
  int ic, i;

  for (ic = yyinput(), i = 0; (ic != c) && (ic != EOF); ic = yyinput(), i++) {

    if (ic == '\n') /* newline should be tested before readescape */ 
      parser.source_line_number++;

    if (ic == '\\') {
      int ignore = 0;
      ic = readescape (&ignore);
      if (ignore) continue;
    }
    yylval.sval[i] = ic;
  }
  yylval.sval[i] = '\0';
}

/*
   This function is copied and modified from c-lex.c of gcc-2.3.3 of
   Free Software Foundation, Inc.

   Read an escape sequence, returning its equivalent as a character,
   or store 1 in *ignore_ptr if it is backslash-newline.
*/

static int readescape(int* ignore_ptr)
{
  register int c = yyinput();
  register int code;
  register unsigned count;
  unsigned firstdig;
  int nonnull;

  switch (c) {
  case 'x':
    code = 0;
    count = 0;
    nonnull = 0;
    while (1) {
      c = yyinput();
      if (!(c >= 'a' && c <= 'f')
	  && !(c >= 'A' && c <= 'F')
	  && !(c >= '0' && c <= '9'))
	{
	  unput(c);
	  parser.source_char_position--;
	  break;
	}
      code *= 16;
      if (c >= 'a' && c <= 'f')
	code += c - 'a' + 10;
      if (c >= 'A' && c <= 'F')
	code += c - 'A' + 10;
      if (c >= '0' && c <= '9')
	code += c - '0';
      if (code != 0 || count != 0)
	{
	  if (count == 0)
	    firstdig = code;
	  count++;
	}
      nonnull = 1;
    }
    if (! nonnull)
      fprintf(stderr, "\\x used with no following hex digits");
    else if (count == 0)
      /* Digits are all 0's.  Ok.  */
      ;
/*    
    else if ((count - 1) * 4 >= TYPE_PRECISION (integer_type_node)
	     || (count > 1
		 && ((1 << (TYPE_PRECISION (integer_type_node) - (count - 1) * 4))
		     <= firstdig)))
      fprintf(stderr, "hex escape out of range");
*/
    return code;

  case '0':  case '1':	case '2':  case '3':	case '4':
  case '5':  case '6':	case '7':
    code = 0;
    count = 0;
    while ((c <= '7') && (c >= '0') && (count++ < 3))
      {
	code = (code * 8) + (c - '0');
	c = yyinput();
      }
    unput(c);
    parser.source_char_position--;
    return code;

  case '\\': case '\'': case '"':
    return c;

  case '\n':
    parser.source_line_number++;
    *ignore_ptr = 1;
    return 0;

  case 'n':
    return '\n';

  case 't':
    return '\t';

  case 'r':
    return '\r';

  case 'f':
    return '\f';

  case 'b':
    return '\b';

  case 'a':
    return '\a';

  case 'v':
    return '\v';

  case 'e':
  case 'E':
    return 033;

  case '?':
    return c;

    /* `\(', etc, are used at beginning of line to avoid confusing Emacs.  */
  case '(':
  case '{':
  case '[':
    return c;
  }
  if (c >= 040 && c < 0177)
    fprintf(stderr, "unknown escape sequence \\%c", c);
  else
    fprintf(stderr, "unknown escape sequence: \\ followed by char code 0x%x", c);
  return c;
}
